/* Copyright (C) 2004 MySQL AB

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */


#include <stdio.h>
#include <glib.h>
#include "myx_public_interface.h"





// ------------------------------------------------------------
// CSV exporter
// ------------------------------------------------------------

typedef struct {
  int dummy;
} MYX_CSV_EXPORTER_INFO;


static MYX_TABLE_EXPORTER_INFO *csv_exporter_init();
static void csv_exporter_free(MYX_TABLE_EXPORTER_INFO *info);
static int csv_exporter_setup(MYX_TABLE_EXPORTER_INFO *self, const char *filename, unsigned int columns_num, MYX_TE_COLUMN *columns);
static void csv_exporter_begin(MYX_TABLE_EXPORTER_INFO *self);
static void csv_exporter_header(MYX_TABLE_EXPORTER_INFO *self, const char *text);
static void csv_exporter_table_header(MYX_TABLE_EXPORTER_INFO *self);
static void csv_exporter_columns(MYX_TABLE_EXPORTER_INFO *self, const char *const*text);
static void csv_exporter_end(MYX_TABLE_EXPORTER_INFO *self);


MYX_TABLE_EXPORTER CSV_TABLE_EXPORTER= {
  "CSV",
  "Generates a CSV (Comma Separated Values) File",
  "Comma Separated Values File",
  "csv",
  csv_exporter_init,
  csv_exporter_free,
  csv_exporter_setup,
  csv_exporter_begin,
  csv_exporter_header,
  csv_exporter_table_header,
  csv_exporter_columns,
  csv_exporter_end,
  {
    {"Separator", ";", "Separator for column values.", MYX_TEOP_STRING},
    {NULL, NULL, NULL, 0}
  }
};


static MYX_TABLE_EXPORTER_INFO *csv_exporter_init()
{
  MYX_TABLE_EXPORTER_INFO *info= g_new0(MYX_TABLE_EXPORTER_INFO, 1);
  unsigned int i;

  info->te= &CSV_TABLE_EXPORTER;

  for (i= 0; info->te->options[i].name; i++);
  info->option_values= g_new0(char*, i);
  for (i= 0; info->te->options[i].name; i++)
    info->option_values[i]= g_strdup(info->te->options[i].value);

  info->priv= g_new0(MYX_CSV_EXPORTER_INFO, 1);

  return info;
}

static void csv_exporter_free(MYX_TABLE_EXPORTER_INFO *self)
{
  unsigned int i;
  for (i= 0; i < self->columns_num; i++)
    g_free(self->columns[i].name);
  g_free(self->columns);

  if (self->file_stream)
    fclose(self->file_stream);

  g_free(self->priv);
  g_free(self);
}


static int csv_exporter_setup(MYX_TABLE_EXPORTER_INFO *self, const char *filename, unsigned int columns_num, MYX_TE_COLUMN *columns)
{
  unsigned int i;

  self->columns_num= columns_num;
  self->columns= g_new(MYX_TE_COLUMN, columns_num);
  for (i= 0; i < columns_num; i++)
  {
    self->columns[i]= columns[i];
    self->columns[i].name= g_strdup(columns[i].name);
  }

  self->file_stream= fopen(filename, "w+");
  if (!self->file_stream)
    return -1;
  return 0;
}


static void csv_exporter_begin(MYX_TABLE_EXPORTER_INFO *self)
{
}


static void csv_exporter_header(MYX_TABLE_EXPORTER_INFO *self, const char *text)
{
  //CVS files must not have a header
  //fprintf(self->file_stream, "# %s\n", text);
}


static void csv_exporter_table_header(MYX_TABLE_EXPORTER_INFO *self)
{
  unsigned int i;
  //fprintf(self->file_stream, "# ");
  for (i= 0; i < self->columns_num; i++)
  {
    char *s;
    if (self->columns[i].name)
      s= str_g_replace(g_strdup(self->columns[i].name), "\"", "\"\"");
    else
      s= NULL;

    if (i > 0)
      fprintf(self->file_stream, "%s\"%s\"", self->option_values[0], s?s:"");
    else
      fprintf(self->file_stream, "\"%s\"", s?s:"");

    g_free(s);
  }

  fprintf(self->file_stream, "\n");
}


static void csv_exporter_columns(MYX_TABLE_EXPORTER_INFO *self, const char *const*text)
{
  unsigned int i;
  for (i= 0; i < self->columns_num; i++)
  {
    if(self->columns[i].ctype == MYX_TE_COL_STRING)
    {
      char *s;
      
      if (text[i])
        s= str_g_replace(g_strdup(text[i]), "\"", "\"\"");
      else
        s= NULL;

      if (i > 0)
        fprintf(self->file_stream, "%s\"%s\"", self->option_values[0], s?s:"");
      else
        fprintf(self->file_stream, "\"%s\"", s?s:"");

      g_free(s);
    }
    else
    {
      if (i > 0)
        fprintf(self->file_stream, "%s%s", self->option_values[0], text[i]);
      else
        fprintf(self->file_stream, "%s", text[i]);
    }
  }
  fprintf(self->file_stream, "\n");
}

static void csv_exporter_end(MYX_TABLE_EXPORTER_INFO *self)
{
}



// ------------------------------------------------------------
// HTML exporter
// ------------------------------------------------------------

typedef struct {
  int row_num;
} MYX_HTML_EXPORTER_INFO;

static MYX_TABLE_EXPORTER_INFO *html_exporter_init();
static void html_exporter_free(MYX_TABLE_EXPORTER_INFO *info);
static int html_exporter_setup(MYX_TABLE_EXPORTER_INFO *self, const char *filename, unsigned int columns_num, MYX_TE_COLUMN *columns);
static void html_exporter_begin(MYX_TABLE_EXPORTER_INFO *self);
static void html_exporter_header(MYX_TABLE_EXPORTER_INFO *self, const char *text);
static void html_exporter_table_header(MYX_TABLE_EXPORTER_INFO *self);
static void html_exporter_columns(MYX_TABLE_EXPORTER_INFO *self, const char *const*text);
static void html_exporter_end(MYX_TABLE_EXPORTER_INFO *self);


MYX_TABLE_EXPORTER HTML_TABLE_EXPORTER= {
  "HTML",
  "Generates a HTML File",
  "HTML File",
  "html",
  html_exporter_init,
  html_exporter_free,
  html_exporter_setup,
  html_exporter_begin,
  html_exporter_header,
  html_exporter_table_header,
  html_exporter_columns,
  html_exporter_end,
  {
    {"Even Row Color", "", "Color for even rows.", MYX_TEOP_COLOR},
    {"Odd Row Color", "", "Color for odd rows.", MYX_TEOP_COLOR},
    {"HTML Table Attributes", "border=1 cellspacing=1 cellpadding=0", "Attributes that will be used in the generated HTML table.", MYX_TEOP_STRING},
    {NULL, NULL, NULL, 0}
  },
};


static MYX_TABLE_EXPORTER_INFO *html_exporter_init()
{
  MYX_TABLE_EXPORTER_INFO *info= g_new0(MYX_TABLE_EXPORTER_INFO, 1);
  unsigned int i;

  info->te= &HTML_TABLE_EXPORTER;

  for (i= 0; info->te->options[i].name; i++);
  info->option_values= g_new0(char*, i);
  for (i= 0; info->te->options[i].name; i++)
    info->option_values[i]= g_strdup(info->te->options[i].value);

  info->priv= g_new0(MYX_HTML_EXPORTER_INFO, 1);

  return info;
}

static void html_exporter_free(MYX_TABLE_EXPORTER_INFO *info)
{
  if (info->file_stream)
    fclose(info->file_stream);
  g_free(info->priv);
  g_free(info);
}


static int html_exporter_setup(MYX_TABLE_EXPORTER_INFO *self, const char *filename, unsigned int columns_num, MYX_TE_COLUMN *columns)
{
  unsigned int i;

  self->columns_num= columns_num;
  self->columns= g_new(MYX_TE_COLUMN, columns_num);
  for (i= 0; i < columns_num; i++)
  {
    self->columns[i]= columns[i];
    self->columns[i].name= g_strdup(columns[i].name);
  }

  self->file_stream= fopen(filename, "w+");
  if (!self->file_stream)
    return -1;
  return 0;
}


static void html_exporter_begin(MYX_TABLE_EXPORTER_INFO *self)
{
}


static void html_exporter_header(MYX_TABLE_EXPORTER_INFO *self, const char *text)
{
  fprintf(self->file_stream, "<html><head><title>%s</title></head>\n", text);
  fprintf(self->file_stream, "<body><h1>%s</h1>\n", text);
}


static void html_exporter_table_header(MYX_TABLE_EXPORTER_INFO *self)
{
  unsigned int i;

  fprintf(self->file_stream, "<table %s><tr>\n", self->option_values[2]);
  for (i= 0; i < self->columns_num; i++)
  {
    fprintf(self->file_stream, "<th>%s</th>", self->columns[i].name);
  }
  fprintf(self->file_stream, "</tr>\n");
  
  ((MYX_HTML_EXPORTER_INFO*)self->priv)->row_num= 0;
}


static void html_exporter_columns(MYX_TABLE_EXPORTER_INFO *self, const char *const*text)
{
  unsigned int i;
  const char *color;
  
  if (((MYX_HTML_EXPORTER_INFO*)self->priv)->row_num%2)
    color= self->option_values[1];
  else
    color= self->option_values[0];

  if (color && *color)
    fprintf(self->file_stream, "<tr bgcolor=\"%s\">\n", color);
  else
    fprintf(self->file_stream, "<tr>\n");
  for (i= 0; i < self->columns_num; i++)
  {
    fprintf(self->file_stream, "<td>%s</td>", text[i]?text[i]:"");
  }
  fprintf(self->file_stream, "</tr>\n");
}


static void html_exporter_end(MYX_TABLE_EXPORTER_INFO *self)
{
  fprintf(self->file_stream, "</table></body></html>");
}


// ------------------------------------------------------------
// XML exporter
// ------------------------------------------------------------

typedef struct {
  char *dtd_filename;
} MYX_XML_EXPORTER_INFO;


static MYX_TABLE_EXPORTER_INFO *xml_exporter_init();
static void xml_exporter_free(MYX_TABLE_EXPORTER_INFO *info);
static int xml_exporter_setup(MYX_TABLE_EXPORTER_INFO *self, const char *filename, unsigned int columns_num, MYX_TE_COLUMN *columns);
static void xml_exporter_begin(MYX_TABLE_EXPORTER_INFO *self);
static void xml_exporter_header(MYX_TABLE_EXPORTER_INFO *self, const char *text);
static void xml_exporter_table_header(MYX_TABLE_EXPORTER_INFO *self);
static void xml_exporter_columns(MYX_TABLE_EXPORTER_INFO *self, const char *const*text);
static void xml_exporter_end(MYX_TABLE_EXPORTER_INFO *self);

MYX_TABLE_EXPORTER XML_TABLE_EXPORTER= {
  "XML",
  "Generates a XML File",
  "XML File",
  "xml",
  xml_exporter_init,
  xml_exporter_free,
  xml_exporter_setup,
  xml_exporter_begin,
  xml_exporter_header,
  xml_exporter_table_header,
  xml_exporter_columns,
  xml_exporter_end,
  {
    {"Generate DTD", "1", "Write the DTD for the generated XML file.", MYX_TEOP_BOOL},
    {"Record Tag", "row", "Name of tag for record delimiter.", MYX_TEOP_STRING},
    {"Root Attributes", "", "Attribute name/value pairs for the root tag.", MYX_TEOP_STRINGLIST},
    {NULL, NULL, NULL, 0}
  }
};


static MYX_TABLE_EXPORTER_INFO *xml_exporter_init()
{
  MYX_TABLE_EXPORTER_INFO *info= g_new0(MYX_TABLE_EXPORTER_INFO, 1);
  unsigned int i;

  info->te= &XML_TABLE_EXPORTER;

  for (i= 0; info->te->options[i].name; i++);
  info->option_values= g_new0(char*, i);
  for (i= 0; info->te->options[i].name; i++)
    info->option_values[i]= g_strdup(info->te->options[i].value);

  info->priv= g_new0(MYX_XML_EXPORTER_INFO, 1);

  return info;
}


static void xml_exporter_free(MYX_TABLE_EXPORTER_INFO *self)
{
  MYX_XML_EXPORTER_INFO* xi= (MYX_XML_EXPORTER_INFO*)self->priv;
  unsigned int i;
  for (i= 0; i < self->columns_num; i++)
    g_free(self->columns[i].name);
  g_free(self->columns);
  g_free(xi->dtd_filename);
  
  if (self->file_stream)
    fclose(self->file_stream);
  g_free(self->priv);
  g_free(self);
}


static int xml_exporter_setup(MYX_TABLE_EXPORTER_INFO *self, const char *filename, unsigned int columns_num, MYX_TE_COLUMN *columns)
{
  MYX_XML_EXPORTER_INFO* xi= (MYX_XML_EXPORTER_INFO*)self->priv;
  char *ptr;  
  unsigned int i;

  self->columns_num= columns_num;
  self->columns= g_new(MYX_TE_COLUMN, columns_num);
  for (i= 0; i < columns_num; i++)
  {
    self->columns[i]= columns[i];
    self->columns[i].name= g_strdup(columns[i].name);
  }
  
  self->file_stream= fopen(filename, "w+");

  ptr= strrchr(filename,'.');
  if (ptr)
    xi->dtd_filename= g_strdup_printf("%.*s.dtd", ptr-filename, filename);
  else
    xi->dtd_filename= g_strdup_printf("%s.dtd", filename);

  return 0;
}


static void xml_exporter_begin(MYX_TABLE_EXPORTER_INFO *self)
{
  fprintf(self->file_stream, "<?xml version=\"1.0\"?>\n");
  fprintf(self->file_stream, "<!DOCTYPE ROOT SYSTEM \".dtd\">\n");
  fprintf(self->file_stream, "<ROOT>\n");
}


static void xml_exporter_header(MYX_TABLE_EXPORTER_INFO *self, const char *text)
{
  
}


static void xml_exporter_table_header(MYX_TABLE_EXPORTER_INFO *self)
{
  unsigned int i;
  MYX_XML_EXPORTER_INFO* xi= (MYX_XML_EXPORTER_INFO*)self->priv;

  // output dtd
  if (atoi(self->option_values[0]))
  {
    FILE *file= fopen(xi->dtd_filename, "w+");
    
    fprintf(file, "<!ELEMENT ROOT ()+>\n");

    fprintf(file, "\t<!ELEMENT (");
    for (i= 0; i < self->columns_num; i++)
    {
      if (i > 0)
        fprintf(file, ",%s", self->columns[i].name);
      else
        fprintf(file, "%s", self->columns[i].name);
    }
    fprintf(file, ")>\n");
    for (i= 0; i < self->columns_num; i++)
    {
      fprintf(file, "\t\t<!ELEMENT %s (#PCDATA)>\n", self->columns[i].name);
    }
    fclose(file);
  }
}


static void xml_exporter_columns(MYX_TABLE_EXPORTER_INFO *self, const char *const*text)
{
  unsigned int i;

  fprintf(self->file_stream, "\t<%s>\n", self->option_values[1]);
  for (i= 0; i < self->columns_num; i++)
  {
    if(text[i])
    {
      char *s= escape_utf8_string_for_xml(g_strdup(text[i]));

      fprintf(self->file_stream, "\t\t<%s>%s</%s>\n", 
              self->columns[i].name, s?s:"", self->columns[i].name);

      g_free(s);
    }
    else
      fprintf(self->file_stream, "\t\t<%s/>\n", 
              self->columns[i].name);

  }
  fprintf(self->file_stream, "\t</%s>\n", self->option_values[1]);
}


static void xml_exporter_end(MYX_TABLE_EXPORTER_INFO *self)
{
  fprintf(self->file_stream, "</ROOT>\n");
}


// ------------------------------------------------------------
// Excel exporter
// ------------------------------------------------------------


typedef struct {
  int dummy;
} MYX_EXCEL_EXPORTER_INFO;


static MYX_TABLE_EXPORTER_INFO *excel_exporter_init();
static void excel_exporter_free(MYX_TABLE_EXPORTER_INFO *info);
static int excel_exporter_setup(MYX_TABLE_EXPORTER_INFO *self, const char *filename, 
                                unsigned int columns_num, MYX_TE_COLUMN *columns);
static void excel_exporter_begin(MYX_TABLE_EXPORTER_INFO *self);
static void excel_exporter_header(MYX_TABLE_EXPORTER_INFO *self, const char *text);
static void excel_exporter_table_header(MYX_TABLE_EXPORTER_INFO *self);
static void excel_exporter_columns(MYX_TABLE_EXPORTER_INFO *self, const char *const*text);
static void excel_exporter_end(MYX_TABLE_EXPORTER_INFO *self);

MYX_TABLE_EXPORTER EXCEL_TABLE_EXPORTER= {
  "Excel",
  "Generates an Excel File",
  "Excel XML File",
  "xls",
  excel_exporter_init,
  excel_exporter_free,
  excel_exporter_setup,
  excel_exporter_begin,
  excel_exporter_header,
  excel_exporter_table_header,
  excel_exporter_columns,
  excel_exporter_end,
  {
    {NULL, NULL, NULL, 0}
  }
};


static MYX_TABLE_EXPORTER_INFO *excel_exporter_init()
{
  MYX_TABLE_EXPORTER_INFO *info= g_new0(MYX_TABLE_EXPORTER_INFO, 1);
  unsigned int i;

  info->te= &EXCEL_TABLE_EXPORTER;

  for (i= 0; info->te->options[i].name; i++);
  info->option_values= g_new0(char*, i);
  for (i= 0; info->te->options[i].name; i++)
    info->option_values[i]= g_strdup(info->te->options[i].value);

  info->priv= g_new0(MYX_EXCEL_EXPORTER_INFO, 1);

  return info;
}


static void excel_exporter_free(MYX_TABLE_EXPORTER_INFO *self)
{
  unsigned int i;

  for (i= 0; i < self->columns_num; i++)
    g_free(self->columns[i].name);
  g_free(self->columns);

  if (self->file_stream)
    fclose(self->file_stream);
  g_free(self->priv);
  g_free(self);
}


static int excel_exporter_setup(MYX_TABLE_EXPORTER_INFO *self, const char *filename, unsigned int columns_num, MYX_TE_COLUMN *columns)
{
  unsigned int i;

  self->columns_num= columns_num;
  self->columns= g_new(MYX_TE_COLUMN, columns_num);
  for (i= 0; i < columns_num; i++)
  {
    self->columns[i]= columns[i];
    self->columns[i].name= g_strdup(columns[i].name);
  }

  self->file_stream= fopen(filename, "w+");
  if (!self->file_stream)
    return -1;

  return 0;
}


static void excel_exporter_begin(MYX_TABLE_EXPORTER_INFO *self)
{
  fprintf(self->file_stream, "<?xml version=\"1.0\"?>\n");
  fprintf(self->file_stream, "<Workbook xmlns=\"urn:schemas-microsoft-com:office:spreadsheet\"\n"
          " xmlns:x=\"urn:schemas-microsoft-com:office:excel\"\n"
          " xmlns:ss=\"urn:schemas-microsoft-com:office:spreadsheet\"\n"
          " xmlns:html=\"http://www.w3.org/TR/REC-html40\">\n");
}


static void excel_exporter_header(MYX_TABLE_EXPORTER_INFO *self, const char *text)
{
  fprintf(self->file_stream, " <Worksheet ss:Name=\"Table1\">\n");
}


static void excel_exporter_table_header(MYX_TABLE_EXPORTER_INFO *self)
{
  unsigned int i;

  fprintf(self->file_stream, "  <Table>\n");

  //Set width for text columns
  for(i= 0; i<self->columns_num; i++)
    if(self->columns[i].ctype == MYX_TE_COL_STRING)
    {
      fprintf(self->file_stream,"   <Column ss:Index=\"%d\" ss:AutoFitWidth=\"0\" ss:Width=\"110\"/>", i+1);
    }

  fprintf(self->file_stream, "   <Row>\n");

  //Add column headers
  for(i= 0; i<self->columns_num; i++)
    fprintf(self->file_stream,"    <Cell><Data ss:Type=\"String\">%s</Data></Cell>\n", 
      self->columns[i].name);

  fprintf(self->file_stream, "   </Row>\n");
}


static void excel_exporter_columns(MYX_TABLE_EXPORTER_INFO *self, const char *const*text)
{
  unsigned int i;
  static char *excel_types[]= {
    "String",//MYX_TE_COL_STRING
    "Number",//MYX_TE_COL_INTEGER
    "Number",//MYX_TE_COL_FLOAT
    "Number"//MYX_TE_COL_BOOL
  };
  
  fprintf(self->file_stream, "   <Row>\n");
  for (i= 0; i < self->columns_num; i++)
  {
    if(text[i])
    {
      char *s= escape_utf8_string_for_xml(g_strdup(text[i]));

      fprintf(self->file_stream, "    <Cell><Data ss:Type=\"%s\">%s</Data></Cell>\n",
              excel_types[self->columns[i].ctype], s);

      g_free(s);
    }
    else
      fprintf(self->file_stream, "    <Cell><Data ss:Type=\"String\"/></Cell>\n");

  }
  fprintf(self->file_stream, "   </Row>\n");
}


static void excel_exporter_end(MYX_TABLE_EXPORTER_INFO *self)
{
  fprintf(self->file_stream, "  </Table>\n");
  fprintf(self->file_stream, " </Worksheet>\n");
  fprintf(self->file_stream, "</Workbook>\n");
}



// ============================================================

static MYX_TABLE_EXPORTER *exporters[]= {
  &CSV_TABLE_EXPORTER,
  &HTML_TABLE_EXPORTER,
  &XML_TABLE_EXPORTER,
  &EXCEL_TABLE_EXPORTER
};


MYX_TABLE_EXPORTER *myx_get_table_exporter(const char *format)
{
  unsigned int i;
  for (i= 0; i < sizeof(exporters)/sizeof(MYX_TABLE_EXPORTER*); i++)
    if (strcmp(format, exporters[i]->name)==0)
      return exporters[i];
  return NULL;
}


MYX_TABLE_EXPORTER_INFO *myx_get_table_exporter_info(const char *format)
{
  const MYX_TABLE_EXPORTER *exporter= myx_get_table_exporter(format);
  
  if (!exporter)
    return NULL;
  
  return (*exporter->init)();
}


int myx_free_table_exporter_info(MYX_TABLE_EXPORTER_INFO *info)
{  
  (*info->te->free)(info);
  return 0;
}


MYX_STRINGLIST *myx_get_table_export_formats()
{
  MYX_STRINGLIST *sl= g_new0(MYX_STRINGLIST, 1);
  unsigned int i= 0;

  sl->strings_num= sizeof(exporters)/sizeof(MYX_TABLE_EXPORTER*);
  sl->strings= g_new0(char*,sl->strings_num);
  for (i= 0; i < sl->strings_num; i++)
    sl->strings[i]= g_strdup(exporters[i]->name);

  return sl;
}

