/* Copyright (C) 2003 MySQL AB

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>

#include "myx_library.h"
#include "myx_network.h"
#include "myx_aux_functions.h"

// needed from my_global.h, cant include it because of conflicts
#ifdef __GNUC__
typedef char    pchar;          /* Mixed prototypes can take char */
typedef char    puchar;         /* Mixed prototypes can take char */
typedef char    pbool;          /* Mixed prototypes can take char */
typedef short   pshort;         /* Mixed prototypes can take short int */
typedef float   pfloat;         /* Mixed prototypes can take float */
#endif

#include <mysql_version.h>
#include <m_ctype.h>

/** @defgroup bridge_to_MySQL_API_private internal stuff
 *  @ingroup bridge_to_MySQL_API */

/** @addtogroup bridge_to_MySQL_API_private
 *  @{ */

/*
 * global variables
 */

enum myx_lib_error error_code;

static char *myx_my_cnf_path= NULL;

///////////////////////////////////////////////////////////////////////////////
/** @brief parameters of mysql_real_connect()
    used for interpretation of MYX_USER_CONNECTION::advanced_options
*//////////////////////////////////////////////////////////////////////////////
typedef struct
{
  MYSQL         * mysql;
  const char    * unix_socket;
  unsigned long   client_flag;
}MYSQL_REAL_CONNECT_PARAMETERS;


typedef struct advanced_option_description ADVANCED_OPTION_DESCRIPTION;

///////////////////////////////////////////////////////////////////////////////
/** @brief function that interpretate value of advanced_option as
           part of parameters for mysql_real_connect() or parameters for
           mysql_options()

    Allowed functions are:
      oi_uint(), oi_client_flag(), oi_unix_socket()
*//////////////////////////////////////////////////////////////////////////////
typedef 
     void advanced_option_interpreter(ADVANCED_OPTION_DESCRIPTION * opt,
                                      const char * value,
                                      MYSQL_REAL_CONNECT_PARAMETERS * pars);

///////////////////////////////////////////////////////////////////////////////
/** @brief description of allowed variant of
           MYX_USER_CONNECTION::advanced_options
*//////////////////////////////////////////////////////////////////////////////
struct advanced_option_description
{
  const char                  * name;      /**< name of option 
                                           (string like "<name_of_option>=") */
  int                           option;    /**< value of option parameter for
                                                         for mysql_options() */
  advanced_option_interpreter * interpret; /**< interpretation method of
                                                                option value */
};

///////////////////////////////////////////////////////////////////////////////
/** @brief interpret value as unsigned int option for mysql_options()
    @param opt description of option (opt->option is the option to set)
    @param value unsigned int value for option
    @param options struct to set option to

    function is used in MYSQL_OPTION_DESCRIPTION struct as value for
      interpret field
*//////////////////////////////////////////////////////////////////////////////
void oi_uint(ADVANCED_OPTION_DESCRIPTION * opt,
             const char * value, MYSQL_REAL_CONNECT_PARAMETERS * options)
{
  int num= atoi(value);
  mysql_options(options->mysql, opt->option, (char*)&num);
}

///////////////////////////////////////////////////////////////////////////////
/** @brief interprets value as flag for client_flag parameter 
            of mysql_real_connect()
    @param opt description of option (opt->option is the option to set)
    @param value unused parameter (only for compatibility with type 
                                   advanced_option_interpreter)
    @param options struct to set option to

    function is used in MYSQL_OPTION_DESCRIPTION struct as value for 
      interpret field..
*//////////////////////////////////////////////////////////////////////////////
void oi_client_flag(ADVANCED_OPTION_DESCRIPTION * opt,
                    const char * value,
                    MYSQL_REAL_CONNECT_PARAMETERS * options)
{
  options->client_flag|= opt->option;
}

///////////////////////////////////////////////////////////////////////////////
/** @brief interprets value as flag for socket_path parameter 
            of mysql_real_connect()
    @param opt description of option
    @param value unused parameter (for compatibility with type 
                                   advanced_option_interpreter only)
    @param options struct to set option to

    function is used in MYSQL_OPTION_DESCRIPTION struct as value for 
      interpret field..
*//////////////////////////////////////////////////////////////////////////////
void oi_unix_socket(ADVANCED_OPTION_DESCRIPTION * opt,
                    const char * value,
                    MYSQL_REAL_CONNECT_PARAMETERS * options)
{
  options->unix_socket= value;
}

///////////////////////////////////////////////////////////////////////////////
/** @brief descriptoins of advanced options                                  */
static ADVANCED_OPTION_DESCRIPTION myx_advanced_options[]=
{
  { "CONNECT_TIMEOUT=", MYSQL_OPT_CONNECT_TIMEOUT, oi_uint        },
  { "COMPRESS=",        CLIENT_COMPRESS,           oi_client_flag },
  { "USE_SSL=",         CLIENT_SSL,                oi_client_flag },
  { "SOCKET_PATH=",     0,                         oi_unix_socket }
};
/** @brief last descriptoin of advanced options in array 
           (is used for iteration)                                           */
ADVANCED_OPTION_DESCRIPTION *last_myx_advanced_option= 
        myx_advanced_options + 
        sizeof(myx_advanced_options)/sizeof(ADVANCED_OPTION_DESCRIPTION);

static MYSQL *static_last_used_mysql= NULL;
static int static_major_version, static_minor_version;

/*
 * functions
 */

///////////////////////////////////////////////////////////////////////////////
/** @brief function for initialization of dll

    initializes pcre_malloc and pcre_free
*//////////////////////////////////////////////////////////////////////////////
#if defined(__WIN__) || defined(_WIN32) || defined(_WIN64)
__declspec(dllexport) BOOL WINAPI DllMain(
    HINSTANCE hinstDLL,  // handle to DLL module
    DWORD fdwReason,     // reason for calling function
    LPVOID lpReserved )  // reserved
{
  // Perform actions based on the reason for calling.
  switch( fdwReason ) 
  { 
    case DLL_PROCESS_ATTACH:
      // Initialize once for each new process.
      // Return FALSE to fail DLL load.
      pcre_malloc= (void*(*)(size_t))g_malloc;
      pcre_free= g_free;
      break;

    case DLL_THREAD_ATTACH:
      // Do thread-specific initialization.
      break;

    case DLL_THREAD_DETACH:
      // Do thread-specific cleanup.
      break;

    case DLL_PROCESS_DETACH:
      // Perform any necessary cleanup.
      break;
  }
  return TRUE;  // Successful DLL_PROCESS_ATTACH.
}
#endif 

///////////////////////////////////////////////////////////////////////////////
/** @brief get version of library
    @ingroup Common_functions
    @return int version number

    uses constant libmysqlx_PUBLIC_INTERFACE_VERSION
*//////////////////////////////////////////////////////////////////////////////
int myx_get_public_interface_version()
{
  return libmysqlx_PUBLIC_INTERFACE_VERSION;
}

///////////////////////////////////////////////////////////////////////////////
/** @brief check if version of mysql is great or equal than given
    @param mysql          connection handler
    @param major_version  given major version of mysql
    @param minor_version  given minor version of mysql
    @return 1 if version of the mysql is greater or equal than given one else 0
*//////////////////////////////////////////////////////////////////////////////
int mysql_version_is_later_or_equal_than(MYSQL *mysql,
                                         int major_version, int minor_version)
{
  int mysql_major_version= myx_get_mysql_major_version(mysql);
  if (mysql_major_version > major_version)
  {
    return 1;
  }
  else if (mysql_major_version < major_version)
  {
    return 0;
  }
  else
  {
    return myx_get_mysql_minor_version(mysql) >= minor_version;
  }
}

///////////////////////////////////////////////////////////////////////////////
/** @brief get path to the configuration file 
            (to use in the myx_mysql_init())
    @ingroup bridge_to_MySQL_API
    @return path to the configure file
    @see myx_set_my_cnf_path(), myx_mysql_init()
*//////////////////////////////////////////////////////////////////////////////
const char *myx_get_my_cnf_path()
{
  return myx_my_cnf_path;
}

///////////////////////////////////////////////////////////////////////////////
/** @brief specify the path to configuration file
           (to use it in the myx_mysql_init())
    @ingroup bridge_to_MySQL_API
    @param path to the configure file
    @return always 0
    @see myx_set_my_cnf_path(), myx_mysql_init()
*//////////////////////////////////////////////////////////////////////////////
int myx_set_my_cnf_path(const char *path)
{
  if (myx_my_cnf_path)
    g_free(myx_my_cnf_path);
  myx_my_cnf_path= path ? g_strdup(path) : NULL;

  return 0;
}

///////////////////////////////////////////////////////////////////////////////
/** @brief create MYSQL handler to work with MySQL API
    @ingroup bridge_to_MySQL_API

    @return prepared MYSQL struct

    calls mysql_init() and reads configuration file (via mysql_options())
     if myx_set_my_cnf_path() was called before
*//////////////////////////////////////////////////////////////////////////////
MYSQL* myx_mysql_init()
{
  MYSQL *mysql= mysql_init(NULL);
  
  if (myx_my_cnf_path)
    mysql_options(mysql, MYSQL_READ_DEFAULT_FILE, myx_my_cnf_path);

  return mysql;
}

///////////////////////////////////////////////////////////////////////////////
/** @brief prepare a connection to a MySQL server
    @ingroup bridge_to_MySQL_API

    @param user_conn connection parameters
    @param mysql     connection handler

    @return Zero if the connection was successful, -1 if the
            connection was unsuccessful

    function parses advanced options of user_conn via myx_advanced_options <BR>
    if version of server is bigger that 4.1 function launches
    "SET CHARACTER SET utf8" query
*//////////////////////////////////////////////////////////////////////////////
int myx_connect_to_instance(MYX_USER_CONNECTION *user_conn, MYSQL *mysql)
{
  MYSQL_REAL_CONNECT_PARAMETERS connect_options= { mysql, 0, 0};  

  /* setup advanced options */
  char **advanced_option= user_conn->advanced_options;
  char **end_advanced_option= advanced_option+user_conn->advanced_options_num;
  for (; advanced_option!=end_advanced_option; advanced_option++)
  {
    char *advanced_value;
    ADVANCED_OPTION_DESCRIPTION * option_description= myx_advanced_options;
    for (; option_description!=last_myx_advanced_option; option_description++)
    {
      if (str_beginswith(*advanced_option, option_description->name) && 
          (advanced_value= strchr(*advanced_option, '=')))
      {
        advanced_value++;
        (*option_description->interpret)(option_description,advanced_value,
                                         &connect_options);
      }
    }
  }

  //This stmt tells the server that the parameters of
  //of mysql_real_connect are in utf8
  //mysql_options(mysql, MYSQL_SET_CHARSET_NAME, "utf8");

  if (!mysql_real_connect(mysql, user_conn->hostname, user_conn->username,
                          user_conn->password, user_conn->schema,
                          user_conn->port, connect_options.unix_socket,
                          connect_options.client_flag | CLIENT_INTERACTIVE))
  {
    return -1;
  }

  //if connected to a server 4.1.x or higher, set charactersets to UTF8
  if (mysql_version_is_later_or_equal_than(mysql,4,1))
  {
    if (mysql_query(mysql, "SET CHARACTER SET utf8"))
      return -1;
  }
  else
  {
    /*char *server_charset= myx_get_server_variable(mysql, "character_set");
    mysql->charset= get_charset_by_csname(server_charset, 0, 0);
    g_free(server_charset);*/
  }

  return 0;
}

///////////////////////////////////////////////////////////////////////////////
/** @brief close a previously opened connection (via myx_mysql_init())
    @ingroup bridge_to_MySQL_API

    @param mysql MYSQL struct to close
    @return always Zero

    deallocates the connection handle pointed to by mysql
    set static_last_used_mysql to NULL if it need
*//////////////////////////////////////////////////////////////////////////////
int myx_mysql_close(MYSQL *mysql)
{
  mysql_close(mysql);
  if (static_last_used_mysql == mysql)
    static_last_used_mysql= NULL;
  return 0;
}

///////////////////////////////////////////////////////////////////////////////
/** @brief use schema (database) by default for connection
    @ingroup bridge_to_MySQL_API

    @param mysql connection handler
    @param schema name of schema to use by default

    @return Zero for success. Non-zero if an error occurred
*//////////////////////////////////////////////////////////////////////////////
int myx_use_schema(MYSQL *mysql, const char *schema)
{
  char *schema_name= myx_convert_utf8_dbstr(mysql, schema);
  int res= mysql_select_db(mysql, schema_name);

  g_free(schema_name);

  return res;
}

///////////////////////////////////////////////////////////////////////////////
/** @brief return the connections default schema (database)
    @ingroup bridge_to_MySQL_API

    @param mysql connection handler

    @return name of default schema that has to be freed with g_free, 
            NULL if an error occurred
*//////////////////////////////////////////////////////////////////////////////
char * myx_get_default_schema(MYSQL *mysql)
{
  return myx_convert_dbstr_utf8(mysql, mysql->db);
}

///////////////////////////////////////////////////////////////////////////////
/** @brief return the connections default schema (database)
    @ingroup bridge_to_MySQL_API

    @param mysql connection handler

    @return the thread id of the given connection
*//////////////////////////////////////////////////////////////////////////////
unsigned long myx_get_thread_id(MYSQL *mysql)
{
  return mysql_thread_id(mysql);
}

///////////////////////////////////////////////////////////////////////////////
/** @brief get mysql_errno() for connection
    @ingroup bridge_to_MySQL_API
    @param mysql connection handler
    @return An error code value for the last mysql_xxx call, if it failed.
            zero means no error occurred.
3A
*//////////////////////////////////////////////////////////////////////////////
int myx_mysql_errno(MYSQL *mysql)
{
  return mysql_errno(mysql);
}

///////////////////////////////////////////////////////////////////////////////
/** @brief get mysql_error() for connection
    @ingroup bridge_to_MySQL_API
    @param mysql connection handler
    @return A null-terminated character string that describes the error.
            An empty string if no error occurred.
	    The returned string must be freed with g_free().
*//////////////////////////////////////////////////////////////////////////////
char *myx_mysql_error(MYSQL *mysql)
{
  return g_strdup(mysql_error(mysql));
}

///////////////////////////////////////////////////////////////////////////////
/** @brief get full version information of mysql server
    @ingroup bridge_to_MySQL_API

    @param mysql connection handler
    @return full version of connection in a string that has to be
            freed with g_free
*//////////////////////////////////////////////////////////////////////////////
char * myx_get_mysql_full_version(MYSQL *mysql)
{
  return g_strdup(mysql_get_server_info(mysql));
}

///////////////////////////////////////////////////////////////////////////////
/** @brief get and cache version of mysql server
    @param mysql connection handler

    calls mysql_get_server_info(), initializes static_major_version and
    static_minor_version, stores mysql into static_last_used_mysql
*//////////////////////////////////////////////////////////////////////////////
void myx_get_mysql_version(MYSQL *mysql)
{
  if (!static_last_used_mysql || static_last_used_mysql != mysql)
  {
    const char *version;
    version= mysql_get_server_info(mysql);
    if ( sscanf(version, "%i.%i.%*s", 
                &static_major_version, &static_minor_version) != 2)
    {
      static_minor_version= 0;
    }
    static_last_used_mysql= mysql;
  }
}

///////////////////////////////////////////////////////////////////////////////
/** @brief get major version of mysql server
    @ingroup bridge_to_MySQL_API

    @param mysql connection handler
    @return major version of connection

    calls myx_get_mysql_version() and returns static_major_version
*//////////////////////////////////////////////////////////////////////////////
int myx_get_mysql_major_version(MYSQL *mysql)
{
  myx_get_mysql_version(mysql);
  return static_major_version;
}

///////////////////////////////////////////////////////////////////////////////
/** @brief get minor version of mysql server
    @ingroup bridge_to_MySQL_API

    @param mysql connection handler
    @return major version of connection

    calls myx_get_mysql_version() and returns static_major_version
*//////////////////////////////////////////////////////////////////////////////
int myx_get_mysql_minor_version(MYSQL *mysql)
{
  myx_get_mysql_version(mysql);
  return static_minor_version;
}

///////////////////////////////////////////////////////////////////////////////
/** @brief get information about the server machine
    @ingroup Host_information_management

    @param user_conn connection parameters
    @param mysql connection handler
    @return prepared information
             - MYX_MACHINE_INFO::version is the string in form 
                                          "MySQL x.x via <name of protocol>"
             - MYX_MACHINE_INFO::network_name is the domain name of machine 
                                                cut by 64 symbols
             - MYX_MACHINE_INFO::IP is the ip string in form "xxx.xxx.xxx.xxx"
             - MYX_MACHINE_INFO::OS is NULL
             - MYX_MACHINE_INFO::hardware is NULL

    calls mysql_get_server_info(), myx_get_ip_as_string(), 
          myx_get_network_name_by_ip()
*//////////////////////////////////////////////////////////////////////////////
MYX_MACHINE_INFO * myx_get_server_info(MYX_USER_CONNECTION *user_conn,
                                       MYSQL *mysql)
{
  MYX_MACHINE_INFO *machine_info= g_malloc(sizeof(MYX_MACHINE_INFO));
  char *add_host_info;

  machine_info->network_name= g_malloc(NETWORK_NAME_BUFFER_LENGTH);
  machine_info->IP=           g_malloc(IP_BUFFER_LENGTH);

#if MYSQL_VERSION_ID >= 50000
  switch (mysql->options.protocol)
  {
  case MYSQL_PROTOCOL_MEMORY: add_host_info= "shared memory"; break;
  case MYSQL_PROTOCOL_SOCKET: add_host_info= "socket";        break;
  case MYSQL_PROTOCOL_PIPE:   add_host_info= "named pipe";    break;
  default:                    add_host_info= "TCP/IP";        break;
  }
#else
  add_host_info= "TCP/IP";
#endif

  machine_info->version=
    g_strdup_printf("MySQL %s via %s",
                    (unsigned char *)mysql_get_server_info(mysql),
                    add_host_info);

  myx_get_ip_as_string(user_conn->hostname, machine_info->IP);
  myx_get_network_name_by_ip(machine_info->IP, machine_info->network_name);

  machine_info->OS= NULL;
  machine_info->hardware= NULL;

  return machine_info;
}

#ifdef __GNUC__
#include <unistd.h>
#endif
///////////////////////////////////////////////////////////////////////////////
/** @brief get information about the client machine
    @ingroup Host_information_management

    @param mysql connection handler
    @return prepared information
             - MYX_MACHINE_INFO::version is the string in form
                                          "MySQL Client Version x.x.x"
             - MYX_MACHINE_INFO::network_name is the domain name of machine
                                                cut by 64 symbols
             - MYX_MACHINE_INFO::IP is the ip string in form "xxx.xxx.xxx.xxx"
             - MYX_MACHINE_INFO::OS is the result of get_local_os_name()
             - MYX_MACHINE_INFO::hardware is the result 
                                           of get_local_hardware_info()

 calls mysql_get_server_info(), myx_get_ip_as_string(),
          myx_get_network_name_by_ip()
*//////////////////////////////////////////////////////////////////////////////
MYX_MACHINE_INFO * myx_get_client_info(MYSQL *mysql)
{
  MYX_MACHINE_INFO *machine_info= g_malloc(sizeof(MYX_MACHINE_INFO));

  machine_info->network_name= g_malloc(NETWORK_NAME_BUFFER_LENGTH);
  machine_info->IP=           g_malloc(IP_BUFFER_LENGTH);

  machine_info->version= 
    g_strdup_printf("MySQL Client Version %i.%i.%i",
                    MYSQL_VERSION_ID/10000,
                    (MYSQL_VERSION_ID/100)%100,MYSQL_VERSION_ID%100);

  gethostname(machine_info->network_name, NETWORK_NAME_BUFFER_LENGTH);
  myx_get_ip_as_string(machine_info->network_name, machine_info->IP);

  machine_info->OS= get_local_os_name();
  machine_info->hardware= get_local_hardware_info();

  return machine_info;
}

///////////////////////////////////////////////////////////////////////////////
/** @brief deallocates memory of MYX_MACHINE_INFO
    @ingroup Host_information_management
    @param machine_info struct to deallocate
    @return always Zero
*//////////////////////////////////////////////////////////////////////////////
int myx_free_pc_info(MYX_MACHINE_INFO *machine_info)
{
  g_free(machine_info->version);
  g_free(machine_info->network_name);
  g_free(machine_info->IP);
  if (machine_info->OS)
    g_free(machine_info->OS);
  if (machine_info->hardware)
    g_free(machine_info->hardware);
  g_free(machine_info);
  return 0;
}


//#if defined(__WIN__) || defined(_WIN32) || defined(_WIN64)
/*
 *----------------------------------------------------------------------
 * myx_match_string
 *
 * SYNOPSIS
 *   Match a string to a search pattern.
 * 
 * DESCRIPTION
 *   Checks whether a text string matches a given search string,
 *   interpreting certain wild characters appropriately.
 * 
 * RETURN VALUE
 * 
 * -1 if the search string is empty
 * 1  if the string matches
 * 0  if the string doesn't match
 *
 * NOTES
 * 
 * Supported wild characters:
 *   %  match anything, any number of chars
 *   _  match a single character
 *
 *----------------------------------------------------------------------
 */
static gchar dummy(gchar c)
{
  return c;
}

int myx_match_string(const char *text, const char *search, int case_sensitive)
{
#define MANY '%'
#define SINGLE '_'

#ifdef __GNUC__
  const char *stack[strlen(text)+1];
#else
  const char **stack= g_malloc(sizeof(char)*((unsigned int)strlen(text)+1));
#endif
  int stackp= 0;
  gchar (*normcase)(gchar)= case_sensitive ? dummy : g_ascii_toupper;

  if (!search || !*search)
    return -1;

  while (*search && *text)
  {
    switch (*search)
    {
     case MANY:
      ++search;
      if (*search != '_')
      {
        gchar norm_search= normcase(*search);
        stack[stackp++]= search;
        while (*text && *search && normcase(*text) != norm_search)
          ++text;
        break;
      }

     case SINGLE:
      ++search;
      ++text;
      break;
      
     default:
      if (normcase(*text) != normcase(*search))
      {
        if (stackp==0)
        {
          return 0;
        }
        else
        {
          // backtrack
          search= stack[--stackp];
          while (*text && *search && normcase(*text) != normcase(*search))
            ++text;
        }
      }
      else
      {
        ++search;
        ++text;
      }
      break;
    }
  }

#ifndef __GNUC__
  g_free((void*)stack);
#endif
  return !*search;
}
//#endif

///////////////////////////////////////////////////////////////////////////////
/** @brief get charset name of server
    @ingroup bridge_to_MySQL_API

    @param mysql connection handler
    @return charset name 
        (mysql->charset->csname if MYSQL_VERSION_ID > 40000 else "latin1")

    returned value isn't supposed to be deallocated..
*//////////////////////////////////////////////////////////////////////////////
const char * myx_get_server_charset_name(MYSQL *mysql)
{
#if MYSQL_VERSION_ID >= 40100
  if (mysql->charset)
    return mysql->charset->csname;
  else
#else
  if (mysql->charset)
    return mysql->charset->name;
  else
#endif
    return "latin1";
}

///////////////////////////////////////////////////////////////////////////////
/** @brief get value of sql-variable from server
    @ingroup bridge_to_MySQL_API

    @param mysql connection handler
    @param name name of variable
    @return value of variable if it exist else NULL

    returned value is supposed to be deallocated via g_free()
*//////////////////////////////////////////////////////////////////////////////
char *myx_get_server_variable(MYSQL *mysql, const char *name)
{
  char *sqlcmd;
  char *value= NULL;
  MYSQL_RES *res;
  MYSQL_ROW row;

  sqlcmd= g_strdup_printf("SHOW VARIABLES LIKE '%s'", name);

  if (!(myx_mysql_query(mysql, sqlcmd)))
  {
    if ((res= mysql_store_result(mysql)))
    {
      if ((row= mysql_fetch_row(res)))
      {
        value= g_strdup(row[1]);
      }
      mysql_free_result(res);
    }
  }

  g_free(sqlcmd);

  return value;
}

///////////////////////////////////////////////////////////////////////////////
/** @brief try to convert str from charset "charsetname" to utf8
    @param str string to convert
    @param charsetname charset to convert from
    @return NULL if conversion was unsucceseful else converted string

    returned value is supposed to be deallocated via g_free()
*//////////////////////////////////////////////////////////////////////////////
static char * try_convert_from_cs_to_utf8(const char * str, const char * charsetname)
{
  GError *error= NULL;
  gchar *utf8= NULL;
  gsize bytes_read= 0;

  utf8= g_convert(str, -1, "UTF-8", charsetname, &bytes_read, NULL, &error);

  //check if it worked
  if (error == NULL)
    return utf8;
  
  g_free(utf8);

  return NULL;
}

///////////////////////////////////////////////////////////////////////////////
/** @brief convert the arguments str which comes from the database
             to UTF8 (if it is not already UTF8)
    @ingroup bridge_to_MySQL_API

    @param mysql connection handler
    @param str   string to convert
    @return converted string

    tryes to convert 
      - from myx_get_server_charset_name()
      - from "ISO-8859-1"

    returned value is supposed to be deallocated via g_free()
*//////////////////////////////////////////////////////////////////////////////
char *myx_convert_dbstr_utf8(MYSQL *mysql, const char* str)
{
  gchar *utf8= NULL;
  const char * server_charset; 

  //If string is NULL return NULL
  if (!str)
    return NULL;

  //Check if string already in UTF8
  if ( g_utf8_validate(str, -1, NULL) == TRUE )
    return g_strdup(str);

  server_charset= iconv_char_name(myx_get_server_charset_name(mysql));

  if ((utf8= try_convert_from_cs_to_utf8(str,server_charset)))
    return utf8;
  if ((utf8= try_convert_from_cs_to_utf8(str,"ISO-8859-1")))
    return utf8;  

  //if we cannot convert, return NULL
  return NULL;
}

///////////////////////////////////////////////////////////////////////////////
/** @brief Converts the argument str which has to be in UTF8
           to a string encoded in the format that the database
           expects.

    @param mysql      connection handler
    @param str        string to convert
    @param str_length length of str
    @param new_length pointer to unsigned int that 
                      is supposed to return the length of converted string to
    @return           a new allocated string in an encoding suitable for the db

    if conversion was unsuccesseful returns g_strndup(str)
          with length cut by 2M <BR>
    returned value is supposed to be deallocated via g_free()
*//////////////////////////////////////////////////////////////////////////////
char *myx_convert_utf8_dbstr_with_length(MYSQL *mysql, const char* str,
                                         int str_length,
                                         unsigned int *new_length)
{
  const int max_returned_length= 2000000;

  //If string is NULL return NULL
  if (!str)
  {
    return NULL;
  }
  else
  {
    GError * error= NULL;
    gchar * res= NULL;
    gsize bytes_read= 0;

    //if connected to a server 4.1.x or higher, return UTF8
    if (mysql_version_is_later_or_equal_than(mysql,4,1))
      return g_memdup(str, str_length+1);

    //Get server charset
    res= g_convert(str, str_length,
                   iconv_char_name(myx_get_server_charset_name(mysql)),
                   "UTF-8", &bytes_read, new_length, &error);

    //check if it worked
    if (!error)
      return res;

    g_free(res);
    *new_length= str_length > max_returned_length
                  ? max_returned_length : str_length;
    return g_strndup(str, *new_length);
  }  
}


///////////////////////////////////////////////////////////////////////////////
/** @brief Converts the argument str which has to be in UTF8
           to a string encoded in the format that the database
           expects.
    @ingroup bridge_to_MySQL_API

    @param mysql connection handler
    @param str   string to convert

    @return      a new allocated string in an encoding suitable for the db

    if conversion was unsuccesseful returns g_strndup(str) 
          with length cut by 2M <BR>
    returned value is supposed to be deallocated via g_free()
    uses myx_convert_utf8_dbstr_with_length()
*//////////////////////////////////////////////////////////////////////////////
char *myx_convert_utf8_dbstr(MYSQL *mysql, const char* str)
{
  unsigned int new_length;
  return myx_convert_utf8_dbstr_with_length(mysql, str,
                                            (unsigned int)strlen(str),
                                            &new_length);
}

///////////////////////////////////////////////////////////////////////////////
/** @brief free string allocated using the library
    @ingroup Common_functions
    @param str string to free

    if (str) g_free(str);
*//////////////////////////////////////////////////////////////////////////////
void myx_free_lib_str(char *str)
{
  if (str)
    g_free(str);
}

///////////////////////////////////////////////////////////////////////////////
/** @brief execute sql query using mysql connection
    @ingroup bridge_to_MySQL_API

    @param mysql connection handler
    @param cmd   text of query to execute

    @return result of mysql_query
      (Zero if the query was successful. Non-zero if an error occurred)

    cmd should be in UTF8, function tryes to myx_convert_utf8_dbstr() it
*//////////////////////////////////////////////////////////////////////////////
int myx_mysql_query(MYSQL *mysql, const char *cmd)
{
  if (mysql_version_is_later_or_equal_than(mysql,4,1))
  {
    return mysql_query(mysql, cmd);
  }
  else
  {
    char *conv_cmd= myx_convert_utf8_dbstr(mysql, cmd);
    if (!conv_cmd)
    {
      return -1;
    }
    else
    {
      int res= mysql_query(mysql, conv_cmd);
      g_free(conv_cmd);
      return res;
    }
  }
}

///////////////////////////////////////////////////////////////////////////////
/** @brief execute sql query with given length using mysql connection
    @ingroup bridge_to_MySQL_API

    @param mysql  connection handler
    @param cmd    text of query to execute
    @param len    length of cmd

    @return result of mysql_query
      (Zero if the query was successful. Non-zero if an error occurred)

    cmd should be in UTF8, function tryes to myx_convert_utf8_dbstr() it
*//////////////////////////////////////////////////////////////////////////////
int myx_mysql_real_query(MYSQL *mysql, const char *cmd, unsigned int len)
{
  //if connected to a server 4.1.x or higher, execute UTF8
  if (mysql_version_is_later_or_equal_than(mysql,4,1))
  {
    //g_assert(g_utf8_validate(cmd, -1, 0));

    return mysql_real_query(mysql, cmd, len);
  }
  else // if connected to a server 4.0.x or lower,
  {    //                                       convert cmd to server's charset
    unsigned int new_length;
    char *conv_cmd=
              myx_convert_utf8_dbstr_with_length(mysql, cmd, len, &new_length);

    if (!conv_cmd)
    {
      return -1;
    }
    else
    {
      int res= mysql_real_query(mysql, conv_cmd, new_length);
      g_free(conv_cmd);
      return res;
    }
  }
}

///////////////////////////////////////////////////////////////////////////////
/** @brief look for string in a MYX_STRINGLIST

    @param stringlist  list of strings to look through
    @param src         string to look for
    @return            index of found string if it exists, else -1
*//////////////////////////////////////////////////////////////////////////////
int myx_str_in_stringlist(MYX_STRINGLIST *stringlist, const char *src)
{
  char ** pos= stringlist->strings;
  char ** end= pos + stringlist->strings_num;

  for(; pos!=end; pos++)
  {
    if(strcmp(*pos, src) == 0)
      return (int)(pos - stringlist->strings);
  }
  return -1;
}


///////////////////////////////////////////////////////////////////////////////
/** @brief free a MYX_STRINGLIST structure

    @param stringlist  list of strings to look through
    @return            index of found string if it exists, else -1
*//////////////////////////////////////////////////////////////////////////////
int myx_free_stringlist(MYX_STRINGLIST *stringlist)
{
  unsigned int i;

  if(!stringlist)
    return 0;

  for (i= 0; i < stringlist->strings_num; i++)
  {
    g_free(stringlist->strings[i]);
  }
  g_free(stringlist->strings);
  g_free(stringlist);

  return 0;
}

///////////////////////////////////////////////////////////////////////////////
/** @brief Check whether an UTF8 string can be converted to another encoding.

    @param string      the utf8 string o check.
    @param encoding    the desired target encoding.
    @return            1 if conversion is possible, 0 if not
*//////////////////////////////////////////////////////////////////////////////
int myx_check_utf8_convertible(const char *str, const char *encoding)
{
  GIConv ic= g_iconv_open(encoding, "utf-8");
  char buffer[256];
  int in_left, out_left, count;
  char *in, *out;
  if (!ic)
  {
    g_message("can't iconv from utf-8 to %s", encoding);
    return -1;
  }

  in_left= (int)strlen(str);
  in= (char *)str;

  while (in_left > 0)
  {
    out= (char*)buffer;
    out_left= sizeof(out);
    count= (int)g_iconv(ic, &in, &in_left, &out, &out_left);
    if ((int)count != 0)
      return 0;
  }

  g_iconv_close(ic);

  return 1;
}

/** @} */ // end of bridge_to_MySQL_API
