/* Copyright (C) 2003,2004 MySQL AB

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include <myx_public_interface.h>
#include <myx_library.h>
#include <myx_aux_functions.h>
#include <string.h>
#include <myx_query.h>
#include <assert.h>
#include <ctype.h>


static const char *SELECT_CLAUSES[]=
{
  "SELECT",
  "FROM",
  "WHERE",
  "GROUP BY",
  "HAVING",
  "ORDER BY",
  "LIMIT",
  NULL
};

int myx_identifier_needs_quotes(const char *name)
{
  unsigned int i;
  unsigned int len= (unsigned int)strlen(name);

  for(i= 0; i<len; i++)
  {
    if(!(
      ((name[i]>='a')&&(name[i]<='z'))||
      ((name[i]>='A')&&(name[i]<='Z'))||
      ((name[i]>='0')&&(name[i]<='9'))||
      (name[i]=='_')||(name[i]=='.')
      ))
    {
      return 1;
    }
  }

  return 0;
}

void myx_query_execute_direct(MYSQL *mysql, const char *sql, MYX_LIB_ERROR *error_code)
{
  *error_code= MYX_NO_ERROR;

  if (myx_mysql_query(mysql, sql))
  {
    *error_code= MYX_SQL_ERROR;
  }
}

MYX_QUERY * query_analyze(MYSQL *mysql, const char *sql)
{
  MYX_QUERY *query= g_malloc0(sizeof(MYX_QUERY));

  query->sql=g_strstrip(g_strdup(sql));
  query->query_type= myx_query_type(query->sql);

  query_build_clauses(query);

  if(query->query_type==MYX_QT_SELECT)
  {
    query_get_query_tables(mysql, query);
    query_get_query_tables_columns(mysql, query);
    query_get_query_columns(query);
  }

  return query;
}

void query_free(MYX_QUERY *query)
{
  unsigned int i, j;
  MYX_Q_TABLE *q_table;
  MYX_Q_TABLE_COLUMN *tbl_col;
  MYX_Q_COLUMN *col;
  MYX_Q_CLAUSE *clause;

  g_free(query->sql);

  for (i= 0; i < query->options_num; i++)
    g_free(query->options[i]);
  g_free(query->options);

  //Free MYX_Q_TABLE tables
  for(i=0;i<query->tables_num;i++)
  {
    q_table= query->tables+i;
    
    if(q_table->catalog)
      g_free(q_table->catalog);
    if(q_table->schema)
      g_free(q_table->schema);
    if(q_table->name)
      g_free(q_table->name);
    if(q_table->alias)
      g_free(q_table->alias);
    if(q_table->fullname)
      g_free(q_table->fullname);

    //Free MYX_Q_TABLE_COLUMN table columns
    for(j=0;j<q_table->columns_num;j++)
    {
      tbl_col= q_table->columns+j;
      g_free(tbl_col->column);
    }

    if(q_table->columns)
      g_free(q_table->columns);

    if(q_table->pk_columns)
      g_free(q_table->pk_columns);
  }
  g_free(query->tables);

  //Free MYX_Q_COLUMN Columns
  for(i=0;i<query->columns_num;i++)
  {
    col= query->columns+i;
    g_free(col->column);
    g_free(col->column_alias);
  }
  g_free(query->columns);

  //Free MYX_Q_CLAUSE clauses
  for(i=0;i<query->clauses_num;i++)
  {
    clause= query->clauses+i;
    g_free(clause->clause);
  }
  if(query->clauses)
    g_free(query->clauses);

  g_free(query);
}


static const char *query_get_select_options(MYX_QUERY *query, const char *sql)
{
  const char *p;
  static char *option_keywords[]={
    "ALL",
      "DISTINCT",
      "DISTINCTROW",
      "HIGH_PRIORITY",
//      "STRAIGHT_JOIN",
      "SQL_SMALL_RESULT",
      "SQL_BIG_RESULT",
      "SQL_BUFFER_RESULT",
      "SQL_CACHE",
      "SQL_NO_CACHE",
      "SQL_CALC_FOUND_ROWS",
      NULL
  };
  
#define ISSPACE(c) ((c)==' ' || (c)=='\t' || (c)=='\r' || (c)=='\n')
  for (;;)
  {
    int i;
    /* skip blanks */
    while (ISSPACE(*sql)) sql++;
    p= sql;
    while (*p && !ISSPACE(*p)) p++;
    for (i= 0; option_keywords[i]; i++)
    {
      if (strncasecmp(sql, option_keywords[i], p-sql)==0)
      {
        query->options_num++;
        query->options= g_realloc(query->options, sizeof(char*)*query->options_num);
        query->options[query->options_num-1]= g_strdup(option_keywords[i]);
        
        sql+= p-sql;
        break;
      }
    }
    if (!option_keywords[i])
      break;
  }
  return sql;
}

void query_build_clause(MYX_QUERY *query, const char **clauses, unsigned int current_clause, char *sql)
{
  unsigned int i;
  const char *p1, *p2;
  MYX_Q_CLAUSE *q_clause;
  MYX_Q_CLAUSE_TYPE clause_type;

  p1= strfindword(sql, clauses[current_clause]);
  if(p1)
  {
    if(query->query_type==MYX_QT_SELECT)
    {
      switch (current_clause)
      {
      case 0:
        clause_type= MYX_QCT_SELECT_CLAUSE;
        query_get_select_options(query, p1+strlen(clauses[current_clause])+1);
        break;
      case 1:
        clause_type= MYX_QCT_FROM_CLAUSE;
        break;
      case 2:
        clause_type= MYX_QCT_WHERE_CLAUSE;
        break;
      case 3:
        clause_type= MYX_QCT_GROUP_CLAUSE;
        break;
      case 4:
        clause_type= MYX_QCT_HAVING_CLAUSE;
        break;
      case 5:
        clause_type= MYX_QCT_ORDER_CLAUSE;
        break;
      case 6:
        clause_type= MYX_QCT_LIMIT_CLAUSE;
        break;
      }
    }

    for(i=current_clause+1;clauses[i];i++)
    {
      p2= strfindword(sql, clauses[i]);
      if(p2)
      {
        if(!query->clauses) {
          query->clauses= g_malloc(sizeof(MYX_Q_CLAUSE)); }
        else { 
          query->clauses= g_realloc(query->clauses, sizeof(MYX_Q_CLAUSE)*(query->clauses_num+1)); }
        q_clause= query->clauses+query->clauses_num;
        q_clause->clause_type= clause_type;
        p1+= strlen(clauses[current_clause])+1;
        q_clause->clause= g_strstrip(g_strndup(query->sql+(p1-sql), (gsize)(p2-p1)));
        q_clause->clause[p2-p1]=0;
        q_clause->start_index= (unsigned int)(p1-sql);
        q_clause->end_index= (unsigned int)(p2-sql);
        if((sql[q_clause->end_index-1]=='\n')||(sql[q_clause->end_index-1]=='\r'))
        {
          q_clause->end_with_linebreak= 1;
        }
        else
        {
          q_clause->end_with_linebreak= 0;
        }
        query->clauses_num++;
        break;
      }
    }

    if(!clauses[i])
    {
      if(!query->clauses) {
        query->clauses= g_malloc(sizeof(MYX_Q_CLAUSE)); }
      else { 
        query->clauses= g_realloc(query->clauses, sizeof(MYX_Q_CLAUSE)*(query->clauses_num+1)); }
      q_clause= query->clauses+query->clauses_num;
      q_clause->clause_type= clause_type;
      q_clause->clause= g_strstrip(g_strdup(query->sql+(p1-sql)+strlen(clauses[current_clause])));
      q_clause->start_index= (unsigned int)(p1-sql);
      q_clause->end_index= (unsigned int)strlen(sql);
      q_clause->end_with_linebreak= 0;
      query->clauses_num++;
    }
  }
}

void query_build_clauses(MYX_QUERY *query)
{
  const char **clauses= NULL;
  unsigned int i;
  char *masked_sql;
//  int maskout= 0;

  if(query->query_type==MYX_QT_SELECT)
  {
    clauses= SELECT_CLAUSES;
  }
  else if(query->query_type==MYX_QT_UPDATE)
  {
  }

  if(clauses)
  {
    //Mask the SQL statment, blend out subqueries, replace them by xxx
    /*masked_sql= g_strdup(query->sql);
    for(i=0;i<strlen(masked_sql);i++)
    {
      if(masked_sql[i]==')')
      {
        maskout--;
      }

      if(maskout>0)
      {
        masked_sql[i]='x';
      }

      if(masked_sql[i]=='(')
      {
        maskout++;
      }
    }*/

    masked_sql= mask_out_string(g_strdup(query->sql), '\'', '\'', 'x');
    masked_sql= mask_out_string_re(masked_sql, "\\(\\s*select\\s", '(', ')', 'x');
    masked_sql= mask_out_string(masked_sql, '(', ')', 'x');
    for(i=0;clauses[i];i++)
    {
      query_build_clause(query, clauses, i, masked_sql);
    }

    g_free(masked_sql);
  }
}

//Be aware that sql needs to be trimmed
MYX_Q_TYPE myx_query_type(const char *sql)
{
  MYX_Q_TYPE q_type= MYX_QT_UNKNOWN;

  if(!(g_ascii_strncasecmp(sql, "SELECT", 6)))
  {
    q_type= MYX_QT_SELECT;
  }
  else if(!(g_ascii_strncasecmp(sql, "UPDATE", 6)))
  {
    q_type= MYX_QT_UPDATE;
  }
  else if(!(g_ascii_strncasecmp(sql, "INSERT", 6)))
  {
    q_type= MYX_QT_INSERT;
  }
  else if(!(g_ascii_strncasecmp(sql, "DELETE", 6)))
  {
    q_type= MYX_QT_DELETE;
  }
  else if(!(g_ascii_strncasecmp(sql, "SHOW", 4)))
  {
    q_type= MYX_QT_SHOW;
  }
  else if(!(g_ascii_strncasecmp(sql, "SET", 3)))
  {
    q_type= MYX_QT_SET;
  }
  else if(!(g_ascii_strncasecmp(sql, "DESC", 4)))
  {
    q_type= MYX_QT_DESCRIBE;
  }
  else if(!strcmp2(sql, ""))
  {
    q_type= MYX_QT_EMPTY;
  }


  return q_type;
}

MYX_QUERY * query_get_query_tables(MYSQL *mysql, MYX_QUERY *query)
{
  pcre *pcre_exp;
  const char *error_str;
  int erroffset, offset;
  int matched[60];
  int rc;

  unsigned int i, k;
  unsigned int dot_count;
  unsigned int mallocated_tables= 5;
  const char *table_name, *alias;
  char **table_name_parts;
  char *table_clause;
  MYX_Q_TABLE *table;

  for(i=0;i<query->clauses_num;i++)
  {
    if(query->clauses[i].clause_type==MYX_QCT_FROM_CLAUSE)
    {
      table_clause= g_strdup_printf(",%s", query->clauses[i].clause);
      query->tables_num= 0;
      query->tables= g_malloc(sizeof(MYX_Q_TABLE)*mallocated_tables);

      //use the following regex to get all tables + alias
      //(,|JOIN|STRAIGHT_JOIN)\s*([\w\.]*)\s*(\w*)
      // 1: join type (be aware that the first table has jointype ,)
      // 2: table
      // 3: table alias

#define RE_NAME "(?:\\w+|`[^`]+`|\"[^\"]+\")"
#define RE_DOTNAME "(?:\\.\\w+|\\.`[^`]+`|\\.\"[^\"]+\")"
      pcre_exp= pcre_compile("(,|JOIN|STRAIGHT_JOIN)\\s*("RE_NAME""RE_DOTNAME"*)\\s*(\\w*)",
                             PCRE_CASELESS, &error_str, &erroffset, NULL);
      
#undef RE_NAME
#undef RE_DOTNAME

      if (pcre_exp)
      {
        offset= 0;
        while((rc= pcre_exec(pcre_exp, NULL, table_clause, (int)strlen(table_clause), offset, 0, matched, sizeof(matched)/sizeof(*matched)))>=0)
        {
          pcre_get_substring(table_clause, matched, rc, 2, &table_name);
          pcre_get_substring(table_clause, matched, rc, 3, &alias);

          table= query->tables+query->tables_num;
          memset(table, 0, sizeof(MYX_Q_TABLE));

          table->fullname= g_strdup(table_name); //has to be quoted already, or command would not work anyway
          table->columns_num= 0;
          table->columns= NULL;
          table->pk_columns_num= 0;
          table->pk_columns= NULL;
          table->relationship_type= 0;
          table->relationship= NULL;

          //Check if alias valid
          if((!(strfindword(alias, "JOIN")))&&
             (!(strfindword(alias, "INNER")))&&
             (!(strfindword(alias, "CROSS")))&&
             (!(strfindword(alias, "STRAIGHT_JOIN")))&&
             (!(strfindword(alias, "LEFT")))&&
             (!(strfindword(alias, "NATURAL")))&&
             (!(strfindword(alias, "RIGHT"))))
          {
            table->alias= g_strdup(alias);
          }
          else
          {
            table->alias= g_strdup("");
          }

          //set table catalog, schema and name
          table_name_parts= g_strsplit(table_name, ".", 3);
          for(k=0;table_name_parts[k];k++);
          dot_count= k-1;

          if(dot_count==2)
          {
            table->catalog= unquote_identifier(g_strdup(table_name_parts[0]));
            table->schema= unquote_identifier(g_strdup(table_name_parts[1]));
            table->name= unquote_identifier(g_strdup(table_name_parts[2]));
          } 
          else if(dot_count==1)
          {
            table->schema= unquote_identifier(g_strdup(table_name_parts[0]));
            table->name= unquote_identifier(g_strdup(table_name_parts[1]));
          } 
          else
          {
            table->schema= g_strdup(mysql->db);
            table->name= unquote_identifier(g_strdup(table_name_parts[0]));
          }

          if(table_name_parts)
            g_strfreev(table_name_parts);


          //allocate more memory if needed
          query->tables_num++;
          if(query->tables_num>=mallocated_tables)
          {
            mallocated_tables+= 5;
            query->tables= g_realloc(query->tables, sizeof(MYX_Q_TABLE)*mallocated_tables);                   
          }

          //Move offset
          offset= matched[1];
          if(strfindword(alias, "JOIN"))
            offset-= 4;
          if(strfindword(alias, "STRAIGHT_JOIN"))
            offset-= 13;

          pcre_free_substring(table_name);
          pcre_free_substring(alias);
        }
        pcre_free(pcre_exp);
      }

      g_free(table_clause);
    }
  }

  return query;
}

MYX_QUERY * query_get_query_tables_columns(MYSQL *mysql, MYX_QUERY *query)
{
  unsigned int i, k;
  MYX_Q_TABLE *q_table;
  MYX_Q_TABLE_COLUMN *q_tbl_col;
  char *sqlcmd;

  MYSQL_RES *res;
  MYSQL_FIELD *fields;
  MYSQL_ROW row;
  unsigned int num_fields;

  char *field;
  unsigned int is_pk_field;

  for(i=0;i<query->tables_num;i++)
  {
    q_table= query->tables+i;

    if(q_table->schema)
      sqlcmd= g_strdup_printf("SHOW FULL COLUMNS FROM `%s`.`%s`", q_table->schema, q_table->name);
    else
      sqlcmd= g_strdup_printf("SHOW FULL COLUMNS FROM %s", q_table->fullname);

    if (!(myx_mysql_query(mysql, sqlcmd))
        && (res= mysql_store_result(mysql)))
    {
      //Depending on the version of the server there might be different columns
      num_fields = mysql_num_fields(res);
      fields = mysql_fetch_fields(res);
      {
        while ((row= mysql_fetch_row(res)))
        {
          char *charset= NULL;
          is_pk_field= 0;
          
          for(k=0;k<num_fields;k++)
          {
            if (strcmp(fields[k].name, "Field")==0)
            {
              field= myx_convert_dbstr_utf8(mysql, row[k]);
            }
            else if (strcmp(fields[k].name, "Key")==0)
            {
              if(!strcmp2(row[k],"PRI"))
                is_pk_field= 1;
            }
            else if (strcmp(fields[k].name, "Collation")==0)
            {
              if (strcmp(row[k],"NULL")!=0)
              {
                char *ptr;
                charset= g_strdup(row[k]);
                ptr= strchr(charset, '_');
                if (ptr)
                  *ptr= 0;
                else
                {
                  g_free(charset);
                  charset= NULL;
                }
              }
            }
          }

          if(q_table->columns==0)
          {
            q_table->columns= g_malloc0(sizeof(MYX_Q_TABLE_COLUMN));
          }
          else
          {
            q_table->columns= g_realloc(q_table->columns, sizeof(MYX_Q_TABLE_COLUMN)*(q_table->columns_num+1));
          }
          q_tbl_col= q_table->columns+q_table->columns_num;
          q_tbl_col->column= g_strdup(field);
          q_tbl_col->is_pk= is_pk_field;
          q_tbl_col->charset= charset;

          q_table->columns_num++;
        }

        g_free(field);
      }

      mysql_free_result(res);
    }
    g_free(sqlcmd);

    if(q_table->schema)
      sqlcmd= g_strdup_printf("SHOW CREATE TABLE `%s`.`%s`", q_table->schema, q_table->name);
    else
      sqlcmd= g_strdup_printf("SHOW CREATE TABLE %s", q_table->fullname);
    if (!(myx_mysql_query(mysql, sqlcmd))
        && (res= mysql_store_result(mysql)))
    {
      //char *ptr;
      pcre *pcre_exp;
      MYSQL_ROW row;
      int matched[12];
      int sc;
      //int offset;
      int erroff;
      const char *err;
      char *masked;
      
      g_free(sqlcmd);
      
      row= mysql_fetch_row(res);
      
      masked= mask_out_string(g_strdup(row[1]), '\'', '\'', 'x');

      pcre_exp= pcre_compile("DEFAULT CHARSET=(\\w+)", 0, &err, &erroff, NULL);
      if ((sc=pcre_exec(pcre_exp, NULL, masked, (int)strlen(masked),
                    0, 0, matched, sizeof(matched)/sizeof(*matched)))>=0)
      {
        const char *cs;

        pcre_get_substring(masked, matched, sc, 1, &cs);

        q_table->charset= g_strdup(cs);
        pcre_free_substring(cs);
      }
      g_free(masked);
      pcre_free(pcre_exp);
    }
  }

  return query;
}

MYX_Q_TABLE * query_get_table_by_alias(MYX_QUERY *query, const char *alias, const char *column)
{
  unsigned int j, k;
  MYX_Q_TABLE *q_table, *q_found_tbl= NULL;
  MYX_Q_TABLE_COLUMN *q_tbl_col;

  for(j=0;j<query->tables_num;j++)
  {
    q_table= query->tables+j;

    if(!(g_utf8_casecollate(alias, q_table->alias)))
    {
      if(column)
      {
        for(k=0;k<q_table->columns_num;k++)
        {
          q_tbl_col= q_table->columns+k;
          if((!(g_utf8_casecollate(column, q_tbl_col->column)))||(!strcmp(column, "*")))
          {
            q_found_tbl= q_table;
            break;
          }
        }
      }
      else
      {
         q_found_tbl= q_table;
      }

      break;
    }
  }

  return q_found_tbl;
}

void query_add_all_table_columns(MYX_QUERY *query, MYX_Q_TABLE *q_found_tbl)
{
  unsigned int j, k;
  MYX_Q_TABLE_COLUMN *q_tbl_col;
  MYX_Q_COLUMN *q_column;

  //if the user selects alias.*
  if(q_found_tbl)
  {
    //Add all columns from the table
    if(!query->columns) {
      assert(sizeof(MYX_Q_COLUMN)*q_found_tbl->columns_num<32*1024);
      query->columns= g_malloc0(sizeof(MYX_Q_COLUMN)*q_found_tbl->columns_num); }
    else {
      query->columns= g_realloc(query->columns, sizeof(MYX_Q_COLUMN)*(query->columns_num+q_found_tbl->columns_num)); }

    for(j=0;j<q_found_tbl->columns_num;j++)
    {
      q_tbl_col= q_found_tbl->columns+j;
      q_column= query->columns+query->columns_num+j;
      q_column->column= g_strdup(q_tbl_col->column);
      q_column->column_alias= g_strdup(q_tbl_col->column);
      q_column->table= q_found_tbl;
    }
    query->columns_num+= q_found_tbl->columns_num;
  }
  else
  {
    //Add all columns from all tables
    for(j=0;j<query->tables_num;j++)
    {
      q_found_tbl= query->tables+j;

      if(!query->columns) {
        query->columns= g_malloc0(sizeof(MYX_Q_COLUMN)*q_found_tbl->columns_num); }
      else {
        query->columns= g_realloc(query->columns, sizeof(MYX_Q_COLUMN)*(query->columns_num+q_found_tbl->columns_num)); }

      for(k=0;k<q_found_tbl->columns_num;k++)
      {
        q_tbl_col= q_found_tbl->columns+k;
        q_column= query->columns+query->columns_num+k;
        q_column->column= g_strdup(q_tbl_col->column);
        q_column->column_alias= g_strdup(q_tbl_col->column);
        q_column->table= q_found_tbl;
      }
      query->columns_num+= q_found_tbl->columns_num;
    }
  }
}

void query_add_table_column(MYX_QUERY *query, MYX_Q_TABLE *q_found_tbl, const char *column, const char *column_alias)
{
  MYX_Q_COLUMN *q_column;

  if(!query->columns) {
    query->columns= g_malloc0(sizeof(MYX_Q_COLUMN)); }
  else {
    query->columns= g_realloc(query->columns, sizeof(MYX_Q_COLUMN)*(query->columns_num+1)); }
  q_column= query->columns+query->columns_num;
  q_column->column= g_strdup(column);
  q_column->column_alias= g_strdup(column_alias);
  if (q_found_tbl)
    q_column->table= q_found_tbl;
  else // if no table found, and there's only one table, use it
  {
    if (query->tables_num==1)
      q_column->table= query->tables;
  }


  query->columns_num++;
}

MYX_QUERY * query_get_query_columns(MYX_QUERY *query)
{
  pcre *pcre_exp;
  const char *error_str;
  int erroffset;
  int matched[60];
  int rc;// offset= 0;
  unsigned int i, j;
  const char *table_alias, *column, *column_alias;
  char *masked_clause;
  char **columns;
  unsigned int columns_num= 0;
  unsigned int pos= 0;

  MYX_Q_TABLE *q_found_tbl;

  for(i=0;i<query->clauses_num;i++)
  {
    if(query->clauses[i].clause_type==MYX_QCT_SELECT_CLAUSE)
    {
      //mask strings and () out, so we have real , seperators left
      masked_clause= mask_out_string(g_strdup(query->clauses[i].clause), '\'', '\'', 'x');
      masked_clause= mask_out_string(masked_clause, '(', ')', 'x');

      //now split the clause by ,
      columns= g_strsplit(masked_clause, ",", 0);

      //now, replace masked columns with real columns
      for(j=0; columns[j]; j++)
      {
        columns[j]= strncpy(columns[j], query->clauses[i].clause+pos, strlen(columns[j]));
        pos+=(unsigned int)strlen(columns[j])+1;

        columns[j]= g_strstrip(columns[j]);
        columns_num++;
      }

      //now use the following regex to get the column infos
      //(((\w*)\.)?(\*))|(((\w*)\.)+(\w+)(\s+as\s+(\w+))?)|((.*)\s+as\s+(\w+))|(.*)
      // 3: table alias (optional)
      // 4: *
      //
      // 7: table alias (optional)
      // 8: column name
      // 10: column alias
      //
      // 12: column or function or subselect
      // 13: column alias
      //
      // 14: column name without table or function or subselect without as

      //pcre_exp= pcre_compile("(((\\w*)\\.)?(\\*))|(((\\w*)\\.)+(\\w+)(\\s+as\\s+(\\w+))?)|((.*)\\s+as\\s+(\\w+))|(.*)",
      //  PCRE_CASELESS, &error_str, &erroffset, NULL);

// an identifier, with optional ` or " quotes, but grouping without the quotes
#define RE_NAME "(?:\\w+|`[^`]+`|\"[^\"]+\")"
      pcre_exp= pcre_compile("((("RE_NAME")\\.)?(\\*))|"
                             "((("RE_NAME")\\.)+("RE_NAME")(\\s+as\\s+(\\w+))?)|"
                             "((.*)\\s+as\\s+(\\w+))|"
                             "(.*)",
                             PCRE_CASELESS, &error_str, &erroffset, NULL);
#undef RE_NAME
      if (pcre_exp)
      {
        //Analyse every column
        for(j=0;j<columns_num;j++)
        {
          table_alias= NULL;
          column= NULL;
          column_alias= NULL;
          
          if((rc= pcre_exec(pcre_exp, NULL, columns[j], (int)strlen(columns[j]), 0, 0, matched, sizeof(matched)/sizeof(*matched)))>=0)
          { 
            char *table_alias_unq;
            char *column_unq;
            
            //table_alias.* or *
            if(rc==5)
            {
              pcre_get_substring(columns[j], matched, rc, 3, &table_alias);
              pcre_get_substring(columns[j], matched, rc, 4, &column);

              table_alias_unq= unquote_identifier(g_strdup(table_alias));

              if(table_alias[0]) {
                q_found_tbl= query_get_table_by_alias(query, table_alias_unq, NULL); }
              else {
                q_found_tbl= NULL; }

              query_add_all_table_columns(query, q_found_tbl);

              g_free(table_alias_unq);
              pcre_free_substring(table_alias);
              pcre_free_substring(column);
            }
            else if(rc==9)
            {
              pcre_get_substring(columns[j], matched, rc, 7, &table_alias);
              pcre_get_substring(columns[j], matched, rc, 8, &column);

              table_alias_unq= unquote_identifier(g_strdup(table_alias));
              column_unq= unquote_identifier(g_strdup(column));
              if(table_alias[0]) {
                q_found_tbl= query_get_table_by_alias(query, table_alias_unq, NULL); }
              else {
                q_found_tbl= NULL; }

              query_add_table_column(query, q_found_tbl, column_unq, column_unq);

              g_free(table_alias_unq);
              g_free(column_unq);
              pcre_free_substring(table_alias);
              pcre_free_substring(column);
            }
            //[table_alias.]column_name as column_alias
            else if(rc==11)
            {
              pcre_get_substring(columns[j], matched, rc, 7, &table_alias);
              pcre_get_substring(columns[j], matched, rc, 8, &column);
              pcre_get_substring(columns[j], matched, rc, 10, &column_alias);

              table_alias_unq= unquote_identifier(g_strdup(table_alias));
              column_unq= unquote_identifier(g_strdup(column));
              
              if(table_alias[0]) {
                q_found_tbl= query_get_table_by_alias(query, table_alias_unq, NULL); }
              else {
                q_found_tbl= NULL; }

              query_add_table_column(query, q_found_tbl, column_unq, column_alias);

              g_free(table_alias_unq);
              g_free(column_unq);
              
              pcre_free_substring(table_alias);
              pcre_free_substring(column);
              pcre_free_substring(column_alias);
            }
            else if(rc==14)
            {
              pcre_get_substring(columns[j], matched, rc, 12, &column);
              pcre_get_substring(columns[j], matched, rc, 13, &column_alias);

              column_unq= unquote_identifier(g_strdup(column));
              
              query_add_table_column(query, NULL, column_unq, column_alias);

              g_free(column_unq);

              pcre_free_substring(column);
              pcre_free_substring(column_alias);
            }
            else if(rc==15)
            {
              pcre_get_substring(columns[j], matched, rc, 14, &column);

              column_unq= unquote_identifier(g_strdup(column));

              query_add_table_column(query, NULL, column_unq, column_unq);

              g_free(column_unq);

              pcre_free_substring(column);
            }
          }
        }

        pcre_free(pcre_exp);
      }

      break;
    }
  }

  return query;
}

unsigned int query_is_editable(MYX_QUERY *query, int enforce_editable)
{
  unsigned int i, j;
  unsigned int editable= 0, col_found;

  MYX_Q_TABLE *q_tbl;
  MYX_Q_TABLE_COLUMN *q_tbl_col;
  MYX_Q_COLUMN *q_col, *q_tbl_pk_col;

  // if a column contains a '(', indicating it uses a function,
  // do not allow it to be editable. check that here because
  // we dont want it to have PK columns added either
  for (i= 0; i < query->columns_num; i++)
  {
    if (strchr(query->columns[i].column, '('))
    {
      return 0;
    }
  }
  
  // if a query contains DISTINCT or DISTINCTROW keywords, then we can't edit it
  for (i= 0; i < query->options_num; i++)
  {
    if (strcasecmp(query->options[i], "DISTINCT")==0
        || strcasecmp(query->options[i], "DISTINCTROW")==0)
      return 0;
  }
  
  //only a query with one table is editable
  if(query->tables_num==1)
  {
    //check if no aggregate functions where used


    //check if pk column was selected (tables doesn't need to be checked since we only got one)
    q_tbl= query->tables;
    for(i=0;i<q_tbl->columns_num;i++)
    {
      q_tbl_col= q_tbl->columns+i;
      if(!q_tbl_col->is_pk)
        continue;

      col_found= 0;      
      for(j=0;j<query->columns_num;j++)
      {
        q_col= query->columns+j;
        
        if(!g_utf8_casecollate(q_tbl_col->column, q_col->column))
        {
          col_found= 1;
          break;
        }
      }

      //If PK is not found and enforce_editable is != 0, 
      //add it but increase pk_columns_added_num so it will not be displayed
      if ( (!col_found) && (enforce_editable) )
      {
        q_col= query_add_column(query, q_tbl, q_tbl_col->column, query->sql, MYX_QCT_SELECT_CLAUSE);
        query->pk_columns_added_num++;

        //add q_column to q_tbl->pk_columns
        if(!q_tbl->pk_columns) {
          q_tbl->pk_columns= g_malloc0(sizeof(MYX_Q_COLUMN)); }
        else {
          q_tbl->pk_columns= g_realloc(q_tbl->pk_columns, sizeof(MYX_Q_COLUMN)*(q_tbl->pk_columns_num+1)); }

        q_tbl_pk_col= (MYX_Q_COLUMN *)(q_tbl->pk_columns)+q_tbl->pk_columns_num;
        *q_tbl_pk_col= *q_col;

        q_tbl->pk_columns_num++;
      }
      else if (col_found)
      {
        //add q_column to q_tbl->pk_columns
        if(!q_tbl->pk_columns) {
          q_tbl->pk_columns= g_malloc0(sizeof(MYX_Q_COLUMN)); }
        else {
          q_tbl->pk_columns= g_realloc(q_tbl->pk_columns, sizeof(MYX_Q_COLUMN)*(q_tbl->pk_columns_num+1)); }

        //*(MYX_Q_COLUMN *)((MYX_Q_COLUMN *)(q_tbl->pk_columns)+q_tbl->pk_columns_num)= *q_col;
        q_tbl_pk_col= (MYX_Q_COLUMN *)(q_tbl->pk_columns)+q_tbl->pk_columns_num;
        *q_tbl_pk_col= *q_col;

        q_tbl->pk_columns_num++;
      }
    }

    editable= q_tbl->pk_columns_num > 0;
  }

  return editable;
}

MYX_Q_COLUMN * query_add_column(MYX_QUERY *query, MYX_Q_TABLE *q_table, const char *column, char *sql, MYX_Q_CLAUSE_TYPE clause_type)
{
  char *new_sql;
  char *add_str;
  unsigned int i;
  MYX_Q_CLAUSE *select_clause= NULL;
  MYX_Q_COLUMN *q_column= NULL;

  for(i=0;i<query->clauses_num;i++)
  {
    if(((MYX_Q_CLAUSE *)(query->clauses+i))->clause_type==MYX_QCT_SELECT_CLAUSE)
    {
      select_clause= query->clauses+i;
      break;
    }
  }

  if(select_clause)
  {
    if(!query->columns) {
      query->columns= g_malloc0(sizeof(MYX_Q_COLUMN)); }
    else {
      query->columns= g_realloc(query->columns, sizeof(MYX_Q_COLUMN)*(query->columns_num+1)); }
    q_column= query->columns+query->columns_num;
    q_column->column= g_strdup(column);
    q_column->column_alias= NULL;
    q_column->table= q_table;

    query->columns_num++;

    add_str= g_strdup_printf(", %s", column);

    new_sql= str_g_insert(query->sql, add_str, select_clause->end_index-1);
    g_free(query->sql);
    query->sql= new_sql;
  }

  return q_column;
}

//Still missing: don't replace parameters in strings
MYX_QUERY * query_set_params(MYX_QUERY *query, MYX_STRINGLIST *parameters)
{
  unsigned int i;
  char *param;
  char *name;
  char *value;

  query->params_num= 0;

  if(!parameters || !parameters->strings_num)
  {
    return query;
  }

  for(i=0;i<parameters->strings_num;i++)
  {
    char *p= parameters->strings[i];
    char *tmp= strchr(p, '=');
	  char *param_value;

    if (!tmp)
      continue;

    assert(tmp-p+1<=256);
    assert(strlen(tmp)+1<=256);
    name= g_malloc((gulong)(sizeof(char)*(tmp-p+1)));
    value= g_malloc((gulong)(sizeof(char)*(strlen(tmp)+1)));

    name_of_str(name, p);
    value_of_str(value, p);

    if(name[0])
    {
      param= g_strdup_printf(":%s", name);
	    param_value= g_strdup_printf("'%s'", value);

      if(g_strstr_len(query->sql, (gssize)strlen(query->sql), param))
      {
        query->params_num++;

        query->sql= str_g_replace(query->sql, param, param_value);
      }

      g_free(param);
	  g_free(param_value);
    }
    g_free(name);
    g_free(value);
  }

  return query;
}


static char *strdup_until_semicolon(const char *sql, int *had_semicolon)
{
  char *s= g_strdup(sql);
  char *end= s+strlen(s);

  while (end > s && isspace(*end) && *end!=';') end--;
  if (*end==';')
  {
    *had_semicolon= 1;
    *end= 0;
  }
  return s;
}

char * myx_query_add_column_to_sql(MYSQL *mysql, const char *catalog, const char *schema, const char *table, const char *column, char *sqls, MYX_Q_CLAUSE_TYPE clause_type, int *cursor_pos)
{
  int has_semicolon= 0;
  char *sql= strdup_until_semicolon(sqls, &has_semicolon);
  MYX_QUERY *query= query_analyze(mysql, sql);
  char *new_sql= NULL;
  char *sql2;
  char *sql3= g_malloc(sizeof(char)*(gulong)(strlen(sql)+strlen(column)+3+7+1)); //malloc enough space for new column, table alias, clause header
  char *sql4= g_malloc(sizeof(char)*(gulong)(strlen(sql)+strlen(column)+3+7+1)); //malloc enough space for new column, table alias, clause header
  char *alias;
  unsigned int i, j;
  MYX_Q_CLAUSE *existing_clause= NULL, *insert_after_clause= NULL;
  MYX_Q_TABLE *q_table= NULL;

  //find table to get alias
  for(i=0;i<query->tables_num;i++)
  {
    if(//(!strcmp2(((MYX_Q_TABLE *)(query->tables+i))->catalog, catalog))&&
      (!strcmp2(((MYX_Q_TABLE *)(query->tables+i))->schema, schema))&&
      (!strcmp2(((MYX_Q_TABLE *)(query->tables+i))->name, table)))
    {
      q_table= query->tables+i;
      break;
    }
  }

  if(!q_table)
  {
    //Add table to query
    new_sql= g_strdup_printf("%s", sql);

    return new_sql;
  }

  //Get table alias
  if(q_table->alias)
  {
    if(q_table->alias[0])
    {
      alias= g_strdup_printf("%s.", q_table->alias);
    }
    else
    {
      alias= g_strdup("");
    }
  }
  else
  {
    alias= g_strdup("");
  }


  //try to find clause
  for(i=0;i<query->clauses_num;i++)
  {
    if(clause_type==((MYX_Q_CLAUSE *)(query->clauses+i))->clause_type)
    {
      existing_clause= query->clauses+i;
      break;
    }
  }

  //If the clause already exists
  if(existing_clause)
  {
    sql3= str_trim(str_left(sql3, sql, existing_clause->end_index));
    sql4= str_trim(strcpy2(sql4, sql+existing_clause->end_index));

    //If SELECT * ..., remove * after first column has been selected
    if(!strcmp("*", existing_clause->clause)&&existing_clause->clause_type==MYX_QCT_SELECT_CLAUSE)
    {
      sql3[strlen(sql3)-2]= 0;
    }

    //when this is the first column in the clause...
    if(!strcmp("*", existing_clause->clause)||!(strcmp("", existing_clause->clause)))
    {
      sql2= g_strdup_printf("%s %s%s", sql3, alias, column);
    }
    //When there is a =,<,> simply append column (WHERE iduser=)
    else if((existing_clause->clause[strlen(existing_clause->clause)-1]=='=')||
      (existing_clause->clause[strlen(existing_clause->clause)-1]=='<')||
      (existing_clause->clause[strlen(existing_clause->clause)-1]=='>'))
    {
      sql2= g_strdup_printf("%s%s%s", sql3, alias, column);
    }
    else
    {
      //In WHERE clause connect by AND
      //!!!TODO: check if the last connection was OR
      if(clause_type==MYX_QCT_WHERE_CLAUSE)
      {
        sql2= g_strdup_printf("%s AND %s%s", sql3, alias, column);
      }
      else
      {
        sql2= g_strdup_printf("%s, %s%s", sql3, alias, column);
      }
    }
    sql2= str_trim(sql2);

    /*//remove windows \r\n, so the returned position is correct
    g_free(sql3);
    sql3= str_g_replace(str_g_replace(g_strdup(sql2), "\r\n", " "), "\n", " ");*/
    sql2= str_trim(sql2);
    *cursor_pos= (int)strlen(sql2); //Set pos so cursor can be placed afterwards

    //Fix for cursor one char off to left in first line
    if((existing_clause->clause_type==MYX_QCT_SELECT_CLAUSE)&&
      (!g_strstr_len(sql2, (gssize)strlen(sql2), "\r")))
      *cursor_pos= (int)strlen(sql2)-1;

    if(existing_clause->end_with_linebreak)
    {
      new_sql= str_trim(g_strdup_printf("%s\n%s", sql2, sql4));
    }
    else
    {
      new_sql= str_trim(g_strdup_printf("%s %s", sql2, sql4));
    }
  }
  else //If not found, add clause
  {
    for(j=clause_type;j>0&&!insert_after_clause;j--)
    {
      for(i=0;i<query->clauses_num;i++)
      {
        if(j==((MYX_Q_CLAUSE *)(query->clauses+i))->clause_type)
        {
          insert_after_clause= query->clauses+i;
          break;
        }
      }
    }

    //There was _no_ clause before this clause
    if(!insert_after_clause)
    {
      sql2= g_strdup_printf("SELECT * FROM %s.%s", schema, table);
      new_sql= myx_query_add_column_to_sql(mysql, catalog, schema, table, column, sql2, clause_type, cursor_pos);
    }
    else
    {
      //Get text before the pos where the net clause will be inserted
      sql3= str_trim(str_left(sql3, sql, insert_after_clause->end_index));
      //Get rest, if not last clause
      if(insert_after_clause->end_index<strlen(sql))
      {
        sql4= str_trim(strcpy2(sql4, sql+insert_after_clause->end_index));
      }
      else
      {
        sql4[0]= 0;
      }

      //Insert new clause
      if(clause_type==MYX_QCT_WHERE_CLAUSE)
        sql2= g_strdup_printf("%s \nWHERE %s", sql3, sql4);
      else if(clause_type==MYX_QCT_GROUP_CLAUSE)
        sql2= g_strdup_printf("%s \nGROUP BY %s", sql3, sql4);
      else if(clause_type==MYX_QCT_HAVING_CLAUSE)
        sql2= g_strdup_printf("%s \nHAVING %s", sql3, sql4);
      else if(clause_type==MYX_QCT_ORDER_CLAUSE)
        sql2= g_strdup_printf("%s \nORDER BY %s", sql3, sql4);
      else if(clause_type==MYX_QCT_SET_CLAUSE)
        sql2= g_strdup_printf("%s SET %s", sql3, sql4);

      new_sql= myx_query_add_column_to_sql(mysql, catalog, schema, table, column, sql2, clause_type, cursor_pos);
    }
  }

  query_free(query);
  g_free(sql2);
  g_free(sql3);
  g_free(sql4);

  /*if (has_semicolon)
    return str_g_append(new_sql,";");
  else*/
  return new_sql;
}

char * myx_query_build_from_clauses(MYX_QUERY *query)
{
  unsigned int i;
  char *sql= NULL;

  for(i= 0; i<query->clauses_num; i++)
  {
    switch (query->clauses[i].clause_type)
    {
      case MYX_QCT_SELECT_CLAUSE:
        sql= str_g_append(sql, "SELECT ");
        break;
      case MYX_QCT_FROM_CLAUSE:
        sql= str_g_append(sql, "FROM ");
        break;
      case MYX_QCT_WHERE_CLAUSE:
        sql= str_g_append(sql, "WHERE ");
        break;
      case MYX_QCT_GROUP_CLAUSE:
        sql= str_g_append(sql, "GROUP BY ");
        break;
      case MYX_QCT_HAVING_CLAUSE:
        sql= str_g_append(sql, "HAVING ");
        break;
      case MYX_QCT_ORDER_CLAUSE:
        sql= str_g_append(sql, "ORDER BY ");
        break;
      case MYX_QCT_LIMIT_CLAUSE:
        sql= str_g_append(sql, "LIMIT ");
        break;
      case MYX_QCT_SET_CLAUSE:
        sql= str_g_append(sql, "SET ");
        break;
      case MYX_QCT_INTO_CLAUSE:
        sql= str_g_append(sql, "INTO ");
        break;
      case MYX_QCT_UPDATE_CLAUSE:
        sql= str_g_append(sql, "UPDATE ");
        break;
    }
    sql= str_g_append(sql, query->clauses[i].clause);
    sql= str_g_append(sql, " ");
  }

  return sql;
}


static int is_alias_free(MYX_QUERY *query, const char *alias)
{
  unsigned int i;
  for (i= 0; i < query->tables_num; i++)
  {
    if (strcmp(query->tables[i].alias, alias)==0)
      return 0;
  }
  return 1;
}


//Returns the alias of the first query-table that is still empty
//if it is not empty, it returns the alias of the new_tablename
char * myx_query_get_table_alias(MYX_QUERY *query, const char *new_tablename)
{
  unsigned int i, j;//, k;
  char alias[8]= {0};
//  int found;
  char *ptr;

  if ((!new_tablename) || (!*new_tablename))
    return g_strdup("");

  // the alias will be 1 char if there's no dupes
  alias[0]= *new_tablename;
  alias[1]= 0;
  if (is_alias_free(query, alias))
    return g_strdup(alias);

  ptr= strchr(new_tablename, '_');
  if (ptr)
  {
    alias[1]= *ptr;
    alias[2]= 0;
    if (is_alias_free(query, alias))
      return g_strdup(alias);

    ptr= strchr(ptr+1, '_');
    if (ptr)
    {
      alias[2]= *ptr;
      alias[3]= 0;
      if (is_alias_free(query, alias))
        return g_strdup(alias);
    }
  }

  for (i= 1; i < strlen(new_tablename); i++)
  {
    alias[1]= new_tablename[i];
    alias[2]= 0;

    if (is_alias_free(query, alias))
      return g_strdup(alias);
  }

  j= 1;
  do 
  {
    sprintf(alias+1, "%i", j++);
  } while (!is_alias_free(query, alias));

  return g_strdup(alias);

#if 0
  for (i= 0; i < query->tables_num; i++)
  {
    if(strcmp2(query->tables[i].name, new_tablename) == 0)
      continue;

    for (j= 0; j < min(strlen(query->tables[i].name), strlen(new_tablename)); j++)
    {
      if(query->tables[i].name[j] != new_tablename[j])
      {
        if(strcmp2(query->tables[i].alias, "") == 0)
          alias[0]= query->tables[i].name[j];
        else
          alias[0]= new_tablename[j];

        if(!myx_identifier_needs_quotes(alias))
        {
          found= FALSE;
          for(k= 0; k<query->tables_num; k++)
            if(strcmp2(query->tables[k].alias, alias) == 0)
            {
              found= TRUE;
              break;
            }

          if(!found)
            break;
          else
            g_free(alias);
        }
        else
          g_free(alias);
      }
    }

    if (alias[0])
    {
      return g_strdup(alias);
    }
    else
    {
      //!!!This needs to handle aliases with length of 2 and more
      return g_strdup(new_tablename);
    }
  }

  alias[0]= new_tablename[0];

  return g_strdup(alias);
#endif
}

char * myx_query_add_table_aliases(MYSQL *mysql, const char *source_sql, const char *new_tablename)
{
  MYX_QUERY *query= query_analyze(mysql, source_sql);
  unsigned int i, j;
  MYX_Q_CLAUSE *clause= NULL;
  MYX_Q_CLAUSE_TYPE clause_type= MYX_QCT_FROM_CLAUSE;

  if(query->query_type == MYX_QT_UPDATE)
    clause_type= MYX_QCT_UPDATE_CLAUSE;


  for (i= 0; i < query->tables_num; i++)
  {
    if(strcmp2(query->tables[i].alias, "") == 0)
    {
      char *new_alias= myx_query_get_table_alias(query, new_tablename);

      if(query->tables[i].alias)
        g_free(query->tables[i].alias);

      query->tables[i].alias= new_alias;

      for (j= 0; j < query->clauses_num; j++)
      {
        if (query->clauses[j].clause_type == clause_type)
        {
          clause= query->clauses+j;
          break;
        }
      }

      if(clause)
      {
        char *full_name;
        char *sql;
        char *tmp, *tmp2;

        if(myx_identifier_needs_quotes(query->tables[i].name))
          full_name= g_strdup_printf("`%s` %s", query->tables[i].name, new_alias);
        else
          full_name= g_strdup_printf("%s %s", query->tables[i].name, new_alias);

        //!!This is buggy... all columns from this table in where, order by, ... need the alias as well
        sql= g_strdup(source_sql);

        tmp= g_strdup_printf(" %s ", query->tables[i].fullname);
        tmp2= g_strdup_printf(" %s ", full_name);
        sql= str_g_replace(sql, tmp, tmp2);
        g_free(tmp);
        g_free(tmp2);

        tmp= g_strdup_printf(" %s", query->tables[i].fullname);
        tmp2= g_strdup_printf(" %s", full_name);
        sql= str_g_replace(sql, tmp, tmp2);
        g_free(tmp);
        g_free(tmp2);

        tmp= g_strdup_printf(" %s\n", query->tables[i].fullname);
        tmp2= g_strdup_printf(" %s\n", full_name);
        sql= str_g_replace(sql, tmp, tmp2);
        g_free(tmp);
        g_free(tmp2);

        tmp= g_strdup_printf(" %s\r", query->tables[i].fullname);
        tmp2= g_strdup_printf(" %s\r", full_name);
        sql= str_g_replace(sql, tmp, tmp2);
        g_free(tmp);
        g_free(tmp2);

        g_free(full_name);
        query_free(query);

        return sql;
      }
    }
  }

  g_free(query);

  return g_strdup(source_sql);
}

char * myx_query_add_table_to_sql(MYSQL *mysql, const char *default_schema,
                                  const char *catalog, const char *schema, const char *table,
                                  const char *sql, MYX_Q_TABLE_ADD_TYPE add_type, int *cursor_pos)
{
  MYX_QUERY *query;
  char *new_sql= NULL;
  char *source_sql= str_trim(myx_query_add_table_aliases(mysql, sql, table));
  unsigned int i;
  MYX_Q_CLAUSE *from_clause= NULL;
  char *specify_schema;
  
  if (!default_schema || strcmp(schema,default_schema)!=0)
  {
    if (myx_identifier_needs_quotes(schema))
      specify_schema= g_strdup_printf("`%s`.", schema);
    else
      specify_schema= g_strdup_printf("%s.", schema);
  }
  else
    specify_schema= g_strdup("");
    

  *cursor_pos= (int)strlen(source_sql);

  //remove terminating ;
  if(source_sql[*cursor_pos-1] == ';')
    source_sql[*cursor_pos-1]= 0;

  query= query_analyze(mysql, source_sql);

  if((add_type == MYX_QTAT_SELECT) || (add_type == MYX_QTAT_SELECT_ADD) || (add_type == MYX_QTAT_SELECT_JOIN) || 
    (add_type == MYX_QTAT_SELECT_LEFT_OUTER_JOIN))
  {
    if((query->query_type == MYX_QT_SELECT) && (add_type != MYX_QTAT_SELECT))
    {
      for (i= 0; i < query->clauses_num; i++)
      {
        if (query->clauses[i].clause_type == MYX_QCT_FROM_CLAUSE)
        {
          from_clause= query->clauses+i;
          break;
        }
      }
      
      if (from_clause)
      {
        MYX_STRINGLIST *rels;
        int clause_end= from_clause->end_index;

        // rewind a few chars if there's a \n in the end of the clause
        while (clause_end>0 && ISSPACE(source_sql[clause_end-1]))
          clause_end--;

        switch (add_type)
        {          
        case MYX_QTAT_SELECT_JOIN:
        case MYX_QTAT_SELECT_LEFT_OUTER_JOIN:
          {
            char *table_alias= myx_query_get_table_alias(query, table);

            for (i= 0; i < query->tables_num; i++)
            {
              //Try master > detail
              rels= myx_check_table_relationship(mysql, schema,
                  table, table_alias,
                  query->tables[i].name, query->tables[i].alias);

              //Try detail > master
              if (!rels)
                rels= myx_check_table_relationship(mysql, schema,
                  query->tables[i].name, query->tables[i].alias,
                  table, table_alias);   

              if (rels)
                break;
            }

            if (rels)
            {
              char *part;
              
              if(add_type == MYX_QTAT_SELECT_LEFT_OUTER_JOIN)
              {
                if(myx_identifier_needs_quotes(table))
                  part= g_strdup_printf("LEFT OUTER JOIN %s`%s` %s ON ", specify_schema, table, table_alias);
                else
                  part= g_strdup_printf("LEFT OUTER JOIN %s%s %s ON ", specify_schema, table, table_alias);

                for (i= 0; i < rels->strings_num; i++)
                {
                  if(i > 0)
                    part= str_g_append(part, ", ");

                  part= str_g_append(part, rels->strings[i]);
                }
                myx_free_stringlist(rels);
                
                new_sql= str_trim(g_strdup_printf("%.*s %s %s", clause_end, source_sql, str_trim(part), source_sql+clause_end));
                *cursor_pos= (int)strlen(new_sql);

                g_free(part);
              }
              else if(MYX_QTAT_SELECT_JOIN)
              {
                MYX_Q_CLAUSE *where_clause= NULL;
                int where_clause_end= from_clause->end_index;

                //Add condition to WHERE clause
                for (i= 0; i < query->clauses_num; i++)
                {
                  if (query->clauses[i].clause_type == MYX_QCT_WHERE_CLAUSE)
                  {
                    where_clause= query->clauses+i;
                    break;
                  }
                }

                if(where_clause)
                {
                  part= g_strdup("AND ");

                  where_clause_end= where_clause->end_index;

                  // rewind a few chars if there's a \n in the end of the clause
                  while (where_clause_end>0 && ISSPACE(source_sql[where_clause_end-1]))
                    where_clause_end--;
                }
                else
                  part= g_strdup(_br "WHERE ");

                for (i= 0; i < rels->strings_num; i++)
                {
                  if(i > 0)
                    part= str_g_append(part, " AND ");

                  if(where_clause)
                  {
                    int last_return;
                    // rewind a few chars if there's a \n in the end of the clause
                    last_return= where_clause->end_index;
                    while (last_return>0 && (source_sql[last_return-1] != '\n'))
                      last_return--;

                    if(where_clause->end_index - last_return + strlen(rels->strings[i]) > MAX_COMMAND_LINE_LENGTH)
                      part= str_g_append(part, _br);
                  }

                  part= str_g_append(part, rels->strings[i]);
                }
                myx_free_stringlist(rels);

                if(where_clause)
                  new_sql= str_trim(g_strdup_printf("%.*s %s %s", where_clause_end, source_sql, part, source_sql+where_clause_end));
                else
                  new_sql= str_trim(g_strdup_printf("%.*s %s %s", clause_end, source_sql, part, source_sql+clause_end));

                g_free(source_sql);
                source_sql= new_sql;

                //Add table to FROM clause
                if(myx_identifier_needs_quotes(table))
                  part= g_strdup_printf(", %s`%s` %s", specify_schema, table, table_alias);
                else
                  part= g_strdup_printf(", %s%s %s", specify_schema, table, table_alias);

                
                new_sql= str_trim(g_strdup_printf("%.*s%s %s", clause_end, source_sql, part, source_sql+clause_end));
                *cursor_pos= (int)strlen(new_sql);

                g_free(part);
              }
              
              g_free(table_alias);

              break;
            }
            else
              g_free(table_alias);
              //fall through to SELECT_ADD
          }

        case MYX_QTAT_SELECT_ADD:
          {
            char *table_alias= myx_query_get_table_alias(query, table);

            if(myx_identifier_needs_quotes(table))
              new_sql= g_strdup_printf("%.*s, %s.`%s` %s%s", clause_end, source_sql, specify_schema, table, table_alias, source_sql+clause_end);
            else
              new_sql= g_strdup_printf("%.*s, %s%s %s%s", clause_end, source_sql, specify_schema, table, table_alias, source_sql+clause_end);
            *cursor_pos= (int)strlen(new_sql);

            g_free(table_alias);
            break;
          }
        }
      }
    }
    else if((query->query_type == MYX_QT_EMPTY) || (add_type == MYX_QTAT_SELECT))
    {
      char *table_alias= g_malloc0(2);
      unsigned int i, len= (unsigned int)strlen(table);

      for(i= 0; i<len; i++)
      {
        table_alias[0]= table[i];
        if(!myx_identifier_needs_quotes(table_alias))
          break;
      }
      if(i == len)
        table_alias[0]= 'a';

      if(myx_identifier_needs_quotes(table))
        new_sql= g_strdup_printf("SELECT * FROM %s`%s` %s", specify_schema, table, table_alias);
      else
        new_sql= g_strdup_printf("SELECT * FROM %s%s %s", specify_schema, table, table_alias);

      *cursor_pos= (int)strlen(new_sql);

      g_free(table_alias);
    }
  }
  else if(add_type == MYX_QTAT_UPDATE)
  {
    if(myx_identifier_needs_quotes(table))
      new_sql= g_strdup_printf("UPDATE %s`%s` SET ", specify_schema, table);
    else
      new_sql= g_strdup_printf("UPDATE %s%s SET ", specify_schema, table);

    *cursor_pos= (int)strlen(new_sql);
  }
  else if(add_type == MYX_QTAT_DELETE)
  {
    if(myx_identifier_needs_quotes(table))
      new_sql= g_strdup_printf("DELETE FROM %s`%s`" _br "WHERE ", specify_schema, table);
    else
      new_sql= g_strdup_printf("DELETE FROM %s%s" _br "WHERE ", specify_schema, table);

    *cursor_pos= (int)strlen(new_sql);
  }
  else if(add_type == MYX_QTAT_INSERT)
  {
    if(myx_identifier_needs_quotes(table))
      new_sql= g_strdup_printf("INSERT INTO %s`%s` VALUES()", specify_schema, table);
    else
      new_sql= g_strdup_printf("INSERT INTO %s%s VALUES()", specify_schema, table);

    *cursor_pos= (int)strlen(new_sql)-2;
  }

  if(!new_sql)
    new_sql= source_sql; //str_g_append(source_sql, ";");
  else
    g_free(source_sql);

  g_free(specify_schema);
  query_free(query);
  
  return new_sql;
}


static void add_relation_to_stringlist(MYX_STRINGLIST *stringlist,
                                       const char *col_table1, const char *alias1,
                                       const char *col_table2, const char *alias2)
{
  char *tmp;
  char *col_table1_q;
  char *col_table2_q;

  if(myx_identifier_needs_quotes(col_table1))
    col_table1_q= g_strdup_printf("`%s`", col_table1);
  else
    col_table1_q= g_strdup(col_table1);

  if(myx_identifier_needs_quotes(col_table2))
    col_table2_q= g_strdup_printf("`%s`", col_table2);
  else
    col_table2_q= g_strdup(col_table2);

  stringlist->strings_num++;
  stringlist->strings= g_realloc(stringlist->strings, sizeof(char *) * stringlist->strings_num);

  if (alias1 && *alias1 && alias2 && *alias2)
    tmp= g_strdup_printf("%s.%s=%s.%s", alias1, col_table1_q, alias2, col_table2_q);
  else if (alias1 && *alias1)
    tmp= g_strdup_printf("%s.%s=%s", alias1, col_table1_q, col_table2_q);
  else if (alias2 && *alias2)
    tmp= g_strdup_printf("%s=%s.%s", col_table1_q, alias2, col_table2_q);
  else
    tmp= g_strdup_printf("%s=%s", col_table1_q, col_table2_q);
  stringlist->strings[stringlist->strings_num-1]= tmp;

  g_free(col_table1_q);
  g_free(col_table2_q);
}


static MYX_NAME_VALUE_PAIR *get_table_columns(MYSQL *mysql, const char *schema,
                                              const char *table,
                                              int *column_num)
{
  MYSQL_RES *res;
  MYSQL_ROW row;
  MYX_NAME_VALUE_PAIR *columns;
  char *sql;
  int name_col= -1, type_col= -1, key_col= -1;

  *column_num= 0;
  sql= g_strdup_printf("show columns from `%s`.`%s`", schema, table);
  if (mysql_query(mysql, sql) == 0 && (res= mysql_store_result(mysql)))
  {
    int i;
    MYSQL_FIELD *fields = mysql_fetch_fields(res);
    for (i= 0; i < (int)mysql_num_fields(res); i++)
    {
      if (strcasecmp(fields[i].name,"Field")==0)
        name_col= i;
      else if (strcasecmp(fields[i].name,"Type")==0)
        type_col= i;
      else if (strcasecmp(fields[i].name,"Key")==0)
        key_col= i;
    }

    *column_num= (int)mysql_num_rows(res);
    columns= g_new(MYX_NAME_VALUE_PAIR, *column_num);
    for (i= 0; i < *column_num; i++)
    {
      row= mysql_fetch_row(res);
      
      columns[i].name= g_strdup(row[name_col]); // column name
      columns[i].value= g_strdup_printf("%c%s", (row[key_col] && strcmp(row[key_col],"PRI")==0)?'*':' ', row[type_col]); // column type + pk info
    }
    mysql_free_result(res);

    g_free(sql);
    return columns;
  }
  g_free(sql);
  return NULL;
}


static int match_key_names(const char *col1_name, const char *table1,
                           const char *col2_name, const char *table2)
{
  char *col;
  
  if (strcasecmp(col1_name, col2_name)==0)
    return 1;
  
  // table1col1 == col2
  col= g_strdup_printf("%s%s", table1, col1_name);
  if (strcasecmp(col, col2_name)==0)
  {
    g_free(col);
    return 1;
  }
  g_free(col);

  // table1_col1 == col2
  col= g_strdup_printf("%s_%s", table1, col1_name);
  if (strcasecmp(col, col2_name)==0)
  {
    g_free(col);
    return 1;
  }
  g_free(col);

  return 0;
}


#define OVECTOR_COUNT 12
MYX_STRINGLIST * myx_check_table_relationship(MYSQL *mysql,
                                              const char *schema,

                                              const char *table1,
                                              const char *alias1,
                                              const char *table2,
                                              const char *alias2)
{
  char *sql, *create_stmt;
  MYSQL_RES *res;
  MYSQL_ROW row;
  MYX_STRINGLIST *stringlist;
  pcre *regex_fk;
  int ovector[OVECTOR_COUNT];
  int start_pos= 0, rc;
  char *engine= NULL;

  sql= g_strdup_printf("show create table `%s`.`%s`", schema, table1);
  if (mysql_query(mysql, sql) || !(res= mysql_store_result(mysql)) )
  {
    g_free(sql);
    return NULL;
  }
  g_free(sql);

  row= mysql_fetch_row(res);
  if (!row)
    return NULL;

  {
    const char *error_str;
    int erroffset;
    regex_fk= pcre_compile("FOREIGN\\s+KEY\\s*\\((\\S+)\\)\\s*REFERENCES\\s+(\\S+)\\s+\\((\\S+)\\)", PCRE_CASELESS|PCRE_DOTALL, &error_str, &erroffset, NULL);

    if (!regex_fk)
      return NULL;
  }
  create_stmt= row[1];

  // get the engine type
  engine= get_value_from_text(create_stmt, (int)strlen(create_stmt), ".*ENGINE=(\\w+).*");

  stringlist= g_malloc0(sizeof(MYX_STRINGLIST));

  if (engine && strcasecmp(engine, "innodb")==0)
  {
    while ((rc= pcre_exec(regex_fk, NULL, create_stmt, (int)strlen(create_stmt), start_pos, 0, ovector, OVECTOR_COUNT) ) > 0)
    {
      const char *col_table1, *table3, *col_table3;
      
      pcre_get_substring(create_stmt, ovector, rc, 1, &col_table1);
      pcre_get_substring(create_stmt, ovector, rc, 2, &table3);
      pcre_get_substring(create_stmt, ovector, rc, 3, &col_table3);
      
      col_table1= unquote_identifier( (char*)col_table1);
      table3= unquote_identifier( (char*)table3);
      col_table3= unquote_identifier( (char*)col_table3);
      
      if (!strcmp(table2, table3))
      {
        add_relation_to_stringlist(stringlist,
                                   col_table1, alias1,
                                   col_table3, alias2);
      }
      
      pcre_free_substring(col_table1);
      pcre_free_substring(table3);
      pcre_free_substring(col_table3);
      
      start_pos= ovector[1];
    }
  }
  else // MyISAM and others
  {
    MYX_NAME_VALUE_PAIR *t1_columns;
    MYX_NAME_VALUE_PAIR *t2_columns;
    int t1_columns_num;
    int t2_columns_num;
    int i, j;

    t1_columns= get_table_columns(mysql, schema, table1, &t1_columns_num);
    t2_columns= get_table_columns(mysql, schema, table2, &t2_columns_num);
    
    if (t1_columns && t2_columns)
    {
      // for each column...
      for (i= 0; i < t1_columns_num; i++)
      {
        // ...that is a PK
        if (t1_columns[i].value[0] != '*') // 1st char indicates whether PK
          continue;
        
        for (j= 0; j < t2_columns_num; j++)
        {
          // if the types match
          if (strcmp(t1_columns[i].value+1, t2_columns[j].value+1)==0)
          {
            // check if the names have similar patterns
            if (match_key_names(t1_columns[i].name, table1,
                                t2_columns[j].name, table2))
            {
              // XXX add regex patterns
              add_relation_to_stringlist(stringlist,
                                         t1_columns[i].name, alias1,
                                         t2_columns[j].name, alias2);
            }
          }
        }
      }
    }

    for (i= 0; i < t1_columns_num; i++)
    {
      g_free(t1_columns[i].name);
      g_free(t1_columns[i].value);
    }
    g_free(t1_columns);
    for (i= 0; i < t2_columns_num; i++)
    {
      g_free(t2_columns[i].name);
      g_free(t2_columns[i].value);
    }
    g_free(t2_columns);
  }

  g_free(engine);
  pcre_free(regex_fk);
  mysql_free_result(res);

  //Only return stringlist if there have been strings added
  if(stringlist->strings_num == 0)
  {
    myx_free_stringlist(stringlist);
    stringlist= NULL;
  }

  return stringlist;
}


char *myx_get_create_table_script(MYSQL *mysql, const char *catalog,
                                  const char *schema, const char *table)
{
  MYSQL_RES *res;
  MYSQL_ROW row;
  char *sql;

  sql= g_strdup_printf("show create table `%s`.`%s`", schema, table);
  if (mysql_query(mysql, sql) || !(res= mysql_store_result(mysql)))
  {
    g_free(sql);
    return NULL;
  }
  g_free(sql);

  row= mysql_fetch_row(res);

  if (row) 
    sql= g_strdup(row[1]);
  else
    sql= NULL;
  mysql_free_result(res);

  return sql;
}


MYX_IMAGE_FORMAT myx_guess_image_format(const char *data, unsigned int length)
{
  static struct {
    char *data;
    unsigned int length;
    MYX_IMAGE_FORMAT format;
  } file_headers[]= {
    {"\x89PNG", 4, MYX_IMG_PNG},
    {"\xd8\xff", 2, MYX_IMG_JPEG},
    {"GIF8", 4, MYX_IMG_GIF},
    {"BM", 2, MYX_IMG_BMP},
    {NULL, 0, MYX_IMG_UNKNOWN}
  };
  unsigned int i;

  for (i= 0; file_headers[i].data; i++)
  {
    if (memcmp(file_headers[i].data, data, min(length, file_headers[i].length))==0)
      return file_headers[i].format;
  }
  return MYX_IMG_UNKNOWN;
}
