/* Copyright (C) 2003,2004 MyQL AB

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include <myx_public_interface.h>
#include <myx_library.h>
#include <myx_aux_functions.h>
#include <myx_recordset.h>
#include <myx_query.h>
#include <string.h>
#include <assert.h>
#include <time.h>

#define ROWS_ALLOCATE_BLOCK_COUNT 100
#define CALLBACK_AFTER_BYTES 1024*10

/* compatibility with <= 4.0.x */
#if MYSQL_VERSION_ID < 40100
#define MYSQL_TYPE_DECIMAL     FIELD_TYPE_DECIMAL
#define MYSQL_TYPE_TINY        FIELD_TYPE_TINY
#define MYSQL_TYPE_SHORT       FIELD_TYPE_SHORT
#define MYSQL_TYPE_LONG        FIELD_TYPE_LONG
#define MYSQL_TYPE_FLOAT       FIELD_TYPE_FLOAT
#define MYSQL_TYPE_DOUBLE      FIELD_TYPE_DOUBLE
#define MYSQL_TYPE_NULL        FIELD_TYPE_NULL
#define MYSQL_TYPE_TIMESTAMP   FIELD_TYPE_TIMESTAMP
#define MYSQL_TYPE_LONGLONG    FIELD_TYPE_LONGLONG
#define MYSQL_TYPE_INT24       FIELD_TYPE_INT24
#define MYSQL_TYPE_DATE        FIELD_TYPE_DATE
#define MYSQL_TYPE_TIME        FIELD_TYPE_TIME
#define MYSQL_TYPE_DATETIME    FIELD_TYPE_DATETIME
#define MYSQL_TYPE_YEAR        FIELD_TYPE_YEAR
#define MYSQL_TYPE_NEWDATE     FIELD_TYPE_NEWDATE
#define MYSQL_TYPE_ENUM        FIELD_TYPE_ENUM
#define MYSQL_TYPE_SET         FIELD_TYPE_SET
#define MYSQL_TYPE_TINY_BLOB   FIELD_TYPE_TINY_BLOB
#define MYSQL_TYPE_MEDIUM_BLOB FIELD_TYPE_MEDIUM_BLOB
#define MYSQL_TYPE_LONG_BLOB   FIELD_TYPE_LONG_BLOB
#define MYSQL_TYPE_BLOB        FIELD_TYPE_BLOB
#define MYSQL_TYPE_VAR_STRING  FIELD_TYPE_VAR_STRING
#define MYSQL_TYPE_STRING      FIELD_TYPE_STRING
#define MYSQL_TYPE_CHAR        FIELD_TYPE_TINY
#define MYSQL_TYPE_INTERVAL    FIELD_TYPE_ENUM
#define MYSQL_TYPE_GEOMETRY    FIELD_TYPE_GEOMETRY
#endif


static int int_compare(const char*, const char*);
static int dummy_compare(const char*, const char*);


static void prepare_value(MYSQL *mysql, const char *value, unsigned int new_value_length, 
                          MYX_RS_COLUMN_TYPE coltype, char *buffer);

/*
 * a list of comparison functions for each type (MYX_RS_COLUMN_TYPE)
 */
int (*SQL_VALUE_COMPARER[])(const char*, const char*)= {
  int_compare, // MYX_RSCT_INTEGER= 0, 
  strcmp, //XXX // MYX_RSCT_FLOAT,
  g_utf8_collate, // MYX_RSCT_STRING, 
  strcmp, // MYX_RSCT_DATE,
  strcmp, // MYX_RSCT_TIME,
  strcmp, // MYX_RSCT_DATETIME,
  dummy_compare, //XXX // MYX_RSCT_BLOB,
  g_utf8_collate, // MYX_RSCT_TEXT,
  g_utf8_collate, //XXX // MYX_RSCT_ENUM,
  g_utf8_collate //XXX // MYX_RSCT_SET
};


/*
  Takes a SQL statement, analyzes and executes it and for selects returns a result set

  SYNOPSIS
    myx_query_execute()
    mysql                Pointer to MYSQL struct
  	sql                  pointer to SQL statement
  	parameters           list of parameters to substitude in the form name=value
  	error_code           errorcode the function returns
    user_data            user defined data that is used in the callback function
  	callback_function    function that is called when a select is executed after CALLBACK_AFTER_BYTES are transfered from the server

  NOTES
    The callback function will be called after CALLBACK_AFTER_BYTES transfered by the server.
	  This only applys to selects.

  TODO
    Make commands except SELECTS work

  RETURN
    MYX_RESULTSET *              when a select is executed successfully
	  NULL                         for all other commands or when an error has occured
	  MYX_LIB_ERROR *error_code    MYX_NO_ERROR on success, other on error
*/

MYX_RESULTSET * myx_query_execute(MYSQL *mysql, const char *sql, int enforce_editable,
  MYX_STRINGLIST *parameters, MYX_LIB_ERROR *error_code, void *user_data,
  int (*progress_row_fetch) (unsigned long current_row_count, unsigned long previous_row_count, MYX_RESULTSET *result_set, void *user_data),
  void (*resultset_realloc_before) (void *user_data),
  void (*resultset_realloc_after) (void *user_data))
{
  MYX_RESULTSET *resultset= g_malloc0(sizeof(MYX_RESULTSET));

  MYSQL_RES *res;
  MYSQL_ROW row;
  MYSQL_FIELD *fields;
  unsigned long *field_lengths;
  unsigned int i, k;
  unsigned long row_count, rows_allocated= 0;
  unsigned long bytes_loaded= 0, previous_callback_row_count= 0;

  MYX_RS_COLUMN *current_column;
  MYX_RS_ROW *current_row;
  MYX_RS_FIELD *current_field;

  *error_code= MYX_NO_ERROR;

  resultset->mysql= mysql;
  resultset->query= query_analyze(mysql, sql);

  //Set parameters
  query_set_params(resultset->query, parameters);

  if ((resultset->query->query_type != MYX_QT_SELECT) &&
    (resultset->query->query_type != MYX_QT_SHOW) &&
    (resultset->query->query_type != MYX_QT_DESCRIBE))
  {
    g_free(resultset);

    myx_query_execute_direct(mysql, sql, error_code);

    return NULL;
  }

  if (resultset->query->query_type == MYX_QT_SELECT)
  {
    resultset->editable= query_is_editable(resultset->query, enforce_editable);
  }

  if (!(myx_mysql_query(mysql, resultset->query->sql)))
  {
    if ((res= mysql_use_result(mysql)))
    {
      //Allocate columns
      resultset->columns_num= mysql_num_fields(res);
      resultset->columns_num_to_display= resultset->columns_num-resultset->query->pk_columns_added_num;
      resultset->columns= g_malloc0(sizeof(MYX_RS_COLUMN) * resultset->columns_num);
      fields = mysql_fetch_fields(res);

      for (i= 0;i<resultset->columns_num;i++)
      {
        current_column= resultset->columns + i;

        //Set column infos
        current_column->name= g_strdup(fields[i].name);
        switch(fields[i].type)
        {
        //To be completed
        case MYSQL_TYPE_TINY:
        case MYSQL_TYPE_SHORT: 
        case MYSQL_TYPE_LONG:
        case MYSQL_TYPE_LONGLONG:
        case MYSQL_TYPE_INT24:
          current_column->column_type= MYX_RSCT_INTEGER;
          break;
        //case MYSQL_TYPE_DECIMAL: is handled as a string
        case MYSQL_TYPE_FLOAT:
        case MYSQL_TYPE_DOUBLE:
          current_column->column_type= MYX_RSCT_FLOAT;
          break;
        case MYSQL_TYPE_BLOB:
        case MYSQL_TYPE_TINY_BLOB:
        case MYSQL_TYPE_MEDIUM_BLOB:
        case MYSQL_TYPE_LONG_BLOB:
          if(fields[i].flags & BINARY_FLAG)
            current_column->column_type= MYX_RSCT_BLOB;
          else
            current_column->column_type= MYX_RSCT_TEXT;
          break;
        case MYSQL_TYPE_DATE:
          current_column->column_type= MYX_RSCT_DATE;
          break;
        case MYSQL_TYPE_TIME:
          current_column->column_type= MYX_RSCT_TIME;
          break;
        case MYSQL_TYPE_DATETIME:
          current_column->column_type= MYX_RSCT_DATETIME;
          break;
        case MYSQL_TYPE_ENUM:
          current_column->column_type= MYX_RSCT_ENUM;
          break;
        case MYSQL_TYPE_SET:
          current_column->column_type= MYX_RSCT_SET;
          break;
        default:
          current_column->column_type= MYX_RSCT_STRING;
          break;
        }
      }

      //Add corresponding tables to columns
      add_table_info_to_rs_columns(resultset);

      //Allocate first block of rows
      resultset->rows= g_malloc0(sizeof(MYX_RS_ROW)*ROWS_ALLOCATE_BLOCK_COUNT);
      rows_allocated= ROWS_ALLOCATE_BLOCK_COUNT;

      row_count= 0;
      while ((row= mysql_fetch_row(res)))
      {
        if (row_count>=rows_allocated)
        {
          if(resultset_realloc_before)
            (* resultset_realloc_before)(user_data);

          resultset->rows= g_realloc(resultset->rows, sizeof(MYX_RS_ROW) * (rows_allocated+ROWS_ALLOCATE_BLOCK_COUNT));
          rows_allocated+= ROWS_ALLOCATE_BLOCK_COUNT;

          if(resultset_realloc_after)
            (* resultset_realloc_after)(user_data);
        }

        //Get field lengths
        field_lengths= mysql_fetch_lengths(res);

        current_row= resultset->rows + row_count;

        current_row->diff= 0;
        
        //Alloc current row
        current_row->fields= g_malloc0(sizeof(MYX_RS_FIELD) * resultset->columns_num);

        for (k= 0;k<resultset->columns_num;k++)
        {
          current_column= resultset->columns + k;
          current_field= current_row->fields + k;

          current_field->value_length= field_lengths[k];
          if (current_field->value_length>0)
          {
            if (current_column->column_type==MYX_RSCT_STRING)
            {
              //This relies on the server returning a NULL termintated string
              current_field->value= myx_convert_dbstr_utf8(mysql, row[k]);
            }
            else
            {
              current_field->value= g_memdup(row[k], current_field->value_length + 1);
              current_field->value[current_field->value_length]= 0;
            }

            bytes_loaded+=current_field->value_length;
          }
        }
        row_count++;

        //Call the callback-function after a fixed number of bytes have been transfered
        if (bytes_loaded>=CALLBACK_AFTER_BYTES)
        {
          resultset->rows_num= row_count;

          if ((*progress_row_fetch)(row_count, previous_callback_row_count, resultset, user_data))
          {
            //Don't free resultset, because the user might want to work with what is fetched so far
            //mysql_free_result(res);
            //myx_query_free_resultset(resultset);
            //return NULL;

            *error_code= MYX_STOP_EXECUTION;
            return resultset;
          }
          else
          {
            previous_callback_row_count= row_count;
            bytes_loaded= 0;
          }

          //Simulate slow connection
          //Sleep(1000);
        }
      }

      resultset->rows_num= row_count; //(unsigned int)mysql_num_rows(res);
      resultset->rows= g_realloc(resultset->rows, sizeof(MYX_RS_ROW)*resultset->rows_num);

      //Call the callback-function one last time to let the program fetch the last records
      if (*error_code != MYX_STOP_EXECUTION)
        (*progress_row_fetch)(resultset->rows_num, previous_callback_row_count, resultset, user_data);

      mysql_free_result(res);
    }
    else
    {
      *error_code= MYX_SQL_ERROR;
      myx_query_free_resultset(resultset);
      return NULL;
    }
  }
  else
  {
    *error_code= MYX_SQL_ERROR;
    myx_query_free_resultset(resultset);
    return NULL;
  }

  return resultset;
}


int myx_query_free_resultset(MYX_RESULTSET *resultset)
{
  MYX_RS_COLUMN *current_column;
  MYX_RS_ROW *current_row;
  MYX_RS_FIELD *current_field;

  unsigned int i, j;

  for(i=0;i<resultset->columns_num;i++)
  {
    current_column= resultset->columns+i;
    g_free(current_column->name);
  }
  g_free(resultset->columns);

  for(i=0;i<resultset->rows_num;i++)
  {
    current_row= resultset->rows+i;

    if (current_row->fields)
    {
      for(j=0; j<resultset->columns_num; j++)
      {
        current_field= current_row->fields+j;
        g_free(current_field->value);
      }
    }
    g_free(current_row->fields);
  }
  g_free(resultset->rows);

  if(resultset->actions)
  {
    for(i=0;i<resultset->actions->actions_num;i++)
    {
      g_free(resultset->actions->actions[i].new_value);
    }
    g_free(resultset->actions->actions);
    g_free(resultset->actions);
  }

  query_free(resultset->query);

  g_free(resultset);

  return 0;
} 

MYX_RS_ACTION * myx_query_create_action(MYX_RS_ACTION_TYPE action_type, unsigned int row, MYX_RS_COLUMN *column, unsigned int new_value_length, const char *new_value)
{
  MYX_RS_ACTION* action= g_malloc0(sizeof(MYX_RS_ACTION));

  action->action= action_type;
  action->row= row;
  action->column= column;
  action->new_value= g_malloc(new_value_length+1);
  memcpy(action->new_value, new_value, new_value_length);
  action->new_value[new_value_length]= 0;
  action->new_value_length= new_value_length;

  return action;
}

MYX_PUBLIC_FUNC int myx_query_free_action(MYX_RS_ACTION *action)
{
  g_free(action);
  
  return 0;
}


int myx_query_add_action(MYX_RESULTSET *resultset, MYX_RS_ACTION *action)
{
  MYX_RS_ACTIONS *actions;
  
  if (!resultset->actions)
    resultset->actions= g_malloc0(sizeof(MYX_RS_ACTIONS));
  
  actions= (MYX_RS_ACTIONS *)resultset->actions;

  actions->actions_num++;
  actions->actions= g_realloc(actions->actions,
                              sizeof(MYX_RS_ACTION)*actions->actions_num);

  memcpy(actions->actions+actions->actions_num-1,
         action, sizeof(MYX_RS_ACTION));

  return actions->actions_num-1;
}


int myx_query_delete_action(MYX_RESULTSET *resultset, unsigned int action)
{
  if (!resultset->actions)
    return -1;

  if (action < resultset->actions->actions_num)
  {
    if (resultset->actions->actions_num > action+1)
      memmove(resultset->actions->actions+action,
              resultset->actions->actions+action+1, 
              sizeof(MYX_RS_ACTION) * (resultset->actions->actions_num-action-1));
    resultset->actions->actions_num--;
    return 0;
  }

  return -1;
}



static void prepare_action_value(MYSQL *mysql, MYX_RS_ACTION *action, char *buffer)
{
  prepare_value(mysql, action->new_value, action->new_value_length,
                action->column->column_type, buffer);
}


static void prepare_value(MYSQL *mysql, const char *value, unsigned int value_length, 
                          MYX_RS_COLUMN_TYPE coltype, char *buffer)
{
  if (value == NULL)
  {
    strcpy(buffer, "NULL");
  }
  else
  {
    /* locate field name */
    switch (coltype)
    {
    case MYX_RSCT_INTEGER:
    case MYX_RSCT_FLOAT:
      memcpy(buffer, value, value_length);
      buffer[value_length]= 0;
      break;

    case MYX_RSCT_ENUM:
    case MYX_RSCT_SET:

    case MYX_RSCT_TIME:
    case MYX_RSCT_DATE:
    case MYX_RSCT_DATETIME:
      g_snprintf(buffer, value_length*2, "'%.*s'", 
                 value_length, value);
      break;

    case MYX_RSCT_STRING:
    case MYX_RSCT_TEXT:
    case MYX_RSCT_BLOB:
      {
        int l;
        strcpy(buffer, "'");
        l= mysql_real_escape_string(mysql, buffer+1, value, value_length);
        strcpy(buffer+1+l, "'");
      }
      break;
    }
  }
}



static char *prepare_pks(MYX_RESULTSET *rs, MYX_RS_ACTION *action)
{
  char **values= NULL;
  MYX_RS_ROW *row= rs->rows + action->row;
  MYX_RS_COLUMN *column= action->column;
  MYX_Q_TABLE *table= column->table;
  MYX_QUERY *query= rs->query;
  unsigned int i;
  char *cond;

  g_assert(rs->columns_num == query->columns_num);

  //if this is not an added row, take value from RS
  if(action->row <= rs->rows_num)
  {
    table= action->column->table;

    // iterate over list of columns
    for (i= 0; i < query->columns_num; i++)
    {
      unsigned int j;
      int is_pk= 0;

      // check if this column is a pk
      for (j= 0; j < table->pk_columns_num; j++)
      {
        if (query->columns[i].table == table &&
            strcmp(query->columns[i].column, table->pk_columns[j].column)==0)
        {
          is_pk= 1;
          break;
        }
      }
      if (is_pk)
      {
        char *tmp= g_malloc(row->fields[i].value_length*2+strlen(rs->columns[i].name)+8);

        sprintf(tmp, "`%s`=", rs->columns[i].name);
      
        prepare_value(rs->mysql, row->fields[i].value, row->fields[i].value_length,
                      rs->columns[i].column_type, tmp+strlen(tmp));

        strlist_g_append(&values, tmp);
      }
    }
  }
  //if this is a new added row, return an empty string
  else
  {
    return g_strdup("");
  }

  if (!values)
    g_error("NO PKS FOUND!");

  // join the condition
  cond= g_strjoinv(" AND ", values);
  g_strfreev(values);
  
  return cond;
}


static void append_error(MYX_RS_ACTION_ERRORS **errors, MYSQL *mysql, 
                         MYX_RS_ACTION *action)
{
  MYX_RS_ACTION_ERROR *error;
  
  if (!*errors)
    *errors= g_malloc0(sizeof(MYX_RS_ACTION_ERRORS));

  
  (*errors)->errors_num++;
  (*errors)->errors= g_realloc((*errors)->errors,
                               sizeof(MYX_RS_ACTION_ERROR)*(*errors)->errors_num);
  
  error= (*errors)->errors + (*errors)->errors_num-1;
  
  error->error= myx_mysql_errno(mysql);
  error->error_text= g_strdup(myx_mysql_error(mysql));
  error->action= action;
}


static void append_charset_error(MYX_RS_ACTION_ERRORS **errors,
                                 const char *charset,
                                 MYX_RS_ACTION *action)
{
  MYX_RS_ACTION_ERROR *error;
  
  if (!*errors)
    *errors= g_malloc0(sizeof(MYX_RS_ACTION_ERRORS));

  (*errors)->errors_num++;
  (*errors)->errors= g_realloc((*errors)->errors,
                               sizeof(MYX_RS_ACTION_ERROR)*(*errors)->errors_num);
  
  error= (*errors)->errors + (*errors)->errors_num-1;
  
  error->error= MYX_UTF8_OUT_OF_RANGE;
  error->error_text= g_strdup_printf("String contains characters invalid in the target character set (%s).",
                                     charset);
  error->action= action;
}


static int check_conversion(GIConv gic, const char *string)
{
  char buffer[256];
  gchar *out, *in= (gchar*)string;
  gsize inb= (gsize)strlen(string), outb;
  size_t sz;
  
  while (inb > 0)
  {
    out= (gchar*)buffer;
    outb= sizeof(buffer);
    sz= g_iconv(gic, &in, &inb, &out, &outb);
    if (sz == (size_t)-1 || sz > 0)
      return 0;
  }
  return 1;
}


MYX_RS_ACTION_ERRORS *myx_query_apply_actions(MYX_RESULTSET *resultset)
{
  MYX_RS_ACTION_ERRORS *errors= NULL;
  unsigned int i;
  unsigned int buffer_size= 1024;
  char *buffer= g_malloc(buffer_size);
  char already_deleted;
  GIConv *gic= g_new0(GIConv, resultset->query->columns_num);
  char **charsets= g_new0(char*, resultset->query->columns_num);

  // create a iconv context for each column, so we can check
  // whether data can be converted to the target charset without losses
  for (i= 0; i < resultset->columns_num; i++)
  {
    MYX_Q_TABLE_COLUMN *tcol= NULL;
    MYX_Q_TABLE *table= resultset->columns[i].table;
    unsigned int j;

    if(!table)
    {
      append_error(&errors, resultset->mysql, resultset->actions->actions+i);

      g_free(buffer);
      return errors;
    }

    // locate the MYX_Q_TABLE_COLUMN for this column
    for (j= 0; j < table->columns_num; j++)
    {
      if (strcmp(table->columns[j].column, resultset->query->columns[i].column)==0)
      {
        tcol= table->columns+j;
        break;
      }
    }
    if (tcol && tcol->charset)
    {
      if (strcasecmp(tcol->charset, "utf-8")==0 || strcasecmp(tcol->charset, "utf8")==0)
        gic[i]= NULL;
      else
      {
        gic[i]= g_iconv_open(tcol->charset, "utf-8");
        charsets[i]= tcol->charset;
      }
    }
    else if (table->charset)
    {
      gic[i]= g_iconv_open(table->charset, "utf-8");
      charsets[i]= table->charset;
    }
  }

  // check whether all data to be sent can be converted to the final
  // charset without problems
  for (i= 0; i < resultset->actions->actions_num; i++)
  {
    unsigned int colnum;
    for (colnum= 0; colnum < resultset->columns_num; colnum++)
    {
      if (resultset->actions->actions[i].column == resultset->columns+colnum)
        break;
    }
    g_return_val_if_fail(colnum < resultset->columns_num, NULL);

    if (gic[colnum] && resultset->actions->actions[i].new_value &&
        !check_conversion(gic[colnum], resultset->actions->actions[i].new_value))
    {
      append_charset_error(&errors, charsets[i], resultset->actions->actions+i);
    }
  }
  for (i= 0; i < resultset->columns_num; i++)
  {
    if (gic[i])
      g_iconv_close(gic[i]);
  }
  g_free(gic);
  g_free(charsets);
  
  // if we have conversion errors, fail
  if (errors)
    return errors;

  i= 0;
  while (i < resultset->actions->actions_num)
  {
    MYX_RS_ACTION *action= resultset->actions->actions+i;

    /* prepare a buffer that will fit the value */
    if (action->new_value_length*2+1 >= buffer_size)
    {
      buffer_size= action->new_value_length*2+1;
      buffer= g_realloc(buffer, buffer_size);
    }

    switch (action->action)
    {
    case MYX_RSA_ADD:
      // go over all ADDs for the same row and commit at once
      {
        unsigned int j;
        char **insert_columns= NULL;
        char **insert_values= NULL;
        char *keys, *values;
        char *query;
        MYX_RS_ACTION_TYPE result;

        //Check if this new row has not been deleted as well
        already_deleted= 0;
        for (j= 0; j < resultset->actions->actions_num; j++)
        {
          MYX_RS_ACTION *action2= resultset->actions->actions+j;

          if(j == i)
            continue;

          if (action2->row == action->row
              && action2->action == MYX_RSA_DELETE
              && action2->column->table == action->column->table)
          {
            action->action= MYX_RSA_APPLIED;
            already_deleted= 1;
            break;
          }
        }
        if(already_deleted)
          continue;

        //Prepare insert
	if (myx_identifier_needs_quotes(action->column->name))
	  strlist_g_append(&insert_columns, g_strdup_printf("`%s`",action->column->name));
	else
	  strlist_g_append(&insert_columns, g_strdup(action->column->name));

        prepare_action_value(resultset->mysql, action, buffer);
        strlist_g_append(&insert_values, g_strdup(buffer));

        //Get all values for the insert
        for (j= i+1; j < resultset->actions->actions_num; j++)
        {
          MYX_RS_ACTION *action2= resultset->actions->actions+j;

          if (action2->row == action->row
              && action2->action == MYX_RSA_ADD
//              && action2->action == action->action
              && action2->column->table == action->column->table)
          {
            strlist_g_append(&insert_columns, g_strdup(action2->column->name));

            prepare_action_value(resultset->mysql, action2, buffer);
            strlist_g_append(&insert_values, g_strdup(buffer));
          }
        }

        // prepare query
        keys= g_strjoinv(",", insert_columns);
        values= g_strjoinv(",", insert_values);

        g_strfreev(insert_columns);
        g_strfreev(insert_values);

        if(action->column->table->schema)
          query= g_strdup_printf("INSERT INTO `%s`.`%s` (%s) VALUES (%s);",
                                action->column->table->schema,
                                action->column->table->name,
                                keys, values);
        else
          query= g_strdup_printf("INSERT INTO `%s` (%s) VALUES (%s);",
                                action->column->table->name,
                                keys, values);
        g_free(keys);
        g_free(values);

        //g_message("QUERY: %s", query);

        // execute query
        if (myx_mysql_query(resultset->mysql, query)!=0)
        {
          result= MYX_RSA_FAILED;
          append_error(&errors, resultset->mysql, action);
        }
        else
          result= MYX_RSA_APPLIED;

        // go over all the involved records and update their status
        for (j= i+1; j < resultset->actions->actions_num; j++)
        {
          MYX_RS_ACTION *action2= resultset->actions->actions+j;

          if (action2->row == action->row
              && action2->action == action->action
              && action2->column->table == action->column->table)
          {
            action2->action= result;
          }
        }
        action->action= result;

        g_free(query);
      }
      break;
    case MYX_RSA_UPDATE:
      // commit one update at a time
      {
        char *query;
        char *pks= NULL;
        
        /* setup primary keys */
        pks= prepare_pks(resultset, action);

        prepare_action_value(resultset->mysql, action, buffer);

        if(action->column->table->schema)
          query= g_strdup_printf("UPDATE `%s`.`%s` SET `%s`=%s WHERE %s;",
                                action->column->table->schema,
                                action->column->table->name,
                                action->column->name, buffer, pks);
        else
          query= g_strdup_printf("UPDATE `%s` SET `%s`=%s WHERE %s;",
                                action->column->table->name,
                                action->column->name, buffer, pks);
        g_free(pks);

        //g_message("QUERY: %s", query);
        
        // execute query
        if (myx_mysql_query(resultset->mysql, query)!=0)
        {
          action->action= MYX_RSA_FAILED;
          append_error(&errors, resultset->mysql, action);
        }
        else
        {
          action->action= MYX_RSA_APPLIED;
        }
        g_free(query);
      }
      break;
    case MYX_RSA_DELETE:
      // just delete it
      {
        char *query;
        char *pks= prepare_pks(resultset, action);

        //do not delete if pks is an empty string,
        //that can occure if a new added row was deleted
        if(*pks)
        {
          if(action->column->table->schema)
            query= g_strdup_printf("DELETE FROM `%s`.`%s` WHERE %s;",
                                  action->column->table->schema,
                                  action->column->table->name,
                                  pks);
          else
            query= g_strdup_printf("DELETE FROM `%s` WHERE %s;",
                                  action->column->table->name,
                                  pks);

          //g_message("QUERY: %s", query);

          // execute query
          if (myx_mysql_query(resultset->mysql, query)!=0)
          {
            action->action= MYX_RSA_FAILED;
            append_error(&errors, resultset->mysql, action);
          }
          else
            action->action= MYX_RSA_APPLIED;
          g_free(query);
        }
        else
        {
          action->action= MYX_RSA_APPLIED;
        }
        g_free(pks);
      }
      break;
      
    case MYX_RSA_APPLIED:
    case MYX_RSA_FAILED:
      // ignore
      break;
      
    case MYX_RSA_DISCARDED:
      // ignore
      break;
    }
    
    i++;
  }
  
  if (buffer)
    g_free(buffer);
  
  return errors;
}

int myx_query_discard_actions(MYX_RESULTSET *resultset)
{
  unsigned int i;
  MYX_RS_ACTION *action;

  if(resultset->actions)
  {
    for(i=0;i<resultset->actions->actions_num;i++)
    {
      action= resultset->actions->actions+i;

      if( (action->action != MYX_RSA_APPLIED) && 
        (action->action != MYX_RSA_FAILED) && 
        (action->action != MYX_RSA_DISCARDED) )
      {
        g_free(action->new_value);
        action->new_value= NULL;
        action->new_value_length= 0;

        action->action= MYX_RSA_DISCARDED;
      }
    }

    /*g_free(resultset->actions->actions);
    g_free(resultset->actions);

    resultset->actions= NULL;*/
  }

  return 0;
}


int myx_query_free_action_errors(MYX_RS_ACTION_ERRORS *errors)
{
  unsigned int i;

  for (i= 0; i < errors->errors_num; i++)
    g_free(errors->errors[i].error_text);
  g_free(errors->errors);
  g_free(errors);
  
  return 0;
}


MYX_RESULTSET * add_table_info_to_rs_columns(MYX_RESULTSET *resultset)
{
  unsigned int i;
  MYX_RS_COLUMN *rs_col;
  MYX_Q_COLUMN *q_col;

  for(i=0;i<MIN(resultset->columns_num, resultset->query->columns_num);i++)
  {
    rs_col= resultset->columns+i;
    q_col= resultset->query->columns+i;

    rs_col->table= q_col->table;
  }

  return resultset;
}


#if 0 //XXX
void prepare_action_list()
{
  // if there's a delete, remove all previous inserts and updates for that row

  // if there are more than 1 update for the same column, remove previous

  // !!! if there's an insert, remove all the inserts for the same column
}
#endif


static int rscmp(const void *a, const void *b, gpointer udata)
{
  MYX_RESULTSET *rs= (MYX_RESULTSET*)udata;
  MYX_RS_ROW *ra= rs->rows + *((unsigned int*)a);
  MYX_RS_ROW *rb= rs->rows + *((unsigned int*)b);
  unsigned int c;
  
  // compare primary keys in order of appearance
  for (c= 0; c < rs->columns_num; c++)
  {
    MYX_Q_TABLE *table= rs->columns[c].table;
    int d;

    if(table)
    {
      if (table->columns[c].is_pk)
      {        
        if (!ra->fields[c].value && !rb->fields[c].value)
          d= 0;
        else if (!ra->fields[c].value && rb->fields[c].value)
          d= -1;
        else if (ra->fields[c].value && !rb->fields[c].value)
          d= 1;
        else
          d= (*SQL_VALUE_COMPARER[rs->columns[c].column_type])(ra->fields[c].value, rb->fields[c].value);
        if (d != 0)
          return d;
      }
    }
  }
  return 0;
}


static int rscmp2(unsigned int *a, unsigned int *b,
                  MYX_RESULTSET *rsa, MYX_RESULTSET *rsb)
{
  MYX_RS_ROW *ra= rsa->rows + *a;
  MYX_RS_ROW *rb= rsb->rows + *b;
  unsigned int c;
  
  // compare primary keys in order of appearance
  for (c= 0; c < rsa->columns_num; c++)
  {
    MYX_Q_TABLE *table= rsa->columns[c].table;
    int d;

    if(table)
    {
      if (table->columns[c].is_pk)
      {      
        if (!ra->fields[c].value && !rb->fields[c].value)
          d= 0;
        else if (!ra->fields[c].value && rb->fields[c].value)
          d= -1;
        else if (ra->fields[c].value && !rb->fields[c].value)
          d= 1;
        else
          d= (*SQL_VALUE_COMPARER[rsa->columns[c].column_type])(ra->fields[c].value, rb->fields[c].value);

        if (d != 0)
          return d;
      }
    }
  }
  return 0;
}


int myx_query_compare_possible(MYX_RESULTSET *ra, MYX_RESULTSET *rb)
{
  unsigned int i;
  
  if (ra->columns_num != rb->columns_num)
    return 0;
  
#define STRCMP(a,b) (!(a) && !(b) ? 0 : strcmp2(a,b))
  
  // compare column names and tables
  for (i= 0; i < ra->columns_num; i++)
  {
    if (!ra->columns[i].table || !rb->columns[i].table)
    {
      if (/*ra->columns[i].table == rb->columns[i].table
          &&*/ STRCMP(ra->columns[i].name, rb->columns[i].name)==0)
        ; // ok
      else
        return 0;
    }
    else if (STRCMP(ra->columns[i].table->catalog,
                    rb->columns[i].table->catalog)!=0
             || STRCMP(ra->columns[i].table->schema,
                       rb->columns[i].table->schema)!=0
             || STRCMP(ra->columns[i].table->name,
                       rb->columns[i].table->name)!=0
             || STRCMP(ra->columns[i].name, rb->columns[i].name)!=0)
      return 0;
  }

  return 1;
}



int myx_query_compare_results(MYX_RESULTSET *ra, MYX_RESULTSET *rb)
{
  unsigned int *indexa;
  unsigned int *indexb;
  MYX_RS_ROW *new_rowsa;
  MYX_RS_ROW *new_rowsb;
  unsigned int ni;
  unsigned int i;
  unsigned int ia, ib;
  unsigned int last_indexa, last_indexb;
  unsigned int total;
  
  g_return_val_if_fail(ra->columns_num == rb->columns_num, -1);
  
  // fill index structs (only non-empty rows)
  indexa= g_malloc(sizeof(unsigned int)*ra->rows_num);
  indexb= g_malloc(sizeof(unsigned int)*rb->rows_num);

  last_indexa= 0;
  for (i= 0; i < ra->rows_num; i++)
  {
    if (ra->rows[i].fields)
      indexa[last_indexa++]= i;
  }
  last_indexb= 0;
  for (i= 0; i < rb->rows_num; i++)
  {
    if (rb->rows[i].fields)
      indexb[last_indexb++]= i;
  }
  // sort index structs

  g_qsort_with_data(indexa, last_indexa, sizeof(unsigned int), rscmp, ra);
  g_qsort_with_data(indexb, last_indexb, sizeof(unsigned int), rscmp, rb);
  
  ia= 0;
  ib= 0;
  ni= 0;

  total= MAX(ra->rows_num, rb->rows_num);
  new_rowsa= g_malloc(sizeof(MYX_RS_ROW)*total);
  new_rowsb= g_malloc(sizeof(MYX_RS_ROW)*total);
  
  // go through resultsets, marking new, missing and different rows
  while (ia < last_indexa || ib < last_indexb)
  {
    int d;
    
    if (ia == last_indexa)
      d= 1;
    else if (ib == last_indexb)
      d= -1;
    else
      d= rscmp2(indexa+ia, indexb+ib, ra, rb);
    
    if (ni >= total)
    {
      total += 256;
      new_rowsa= g_realloc(new_rowsa, sizeof(MYX_RS_ROW)*total);
      new_rowsb= g_realloc(new_rowsb, sizeof(MYX_RS_ROW)*total);
    }

    if (d < 0)
    {
      if (ia < last_indexa)
      {
        // this row only exists in A
        new_rowsa[ni]= ra->rows[indexa[ia]];
        memset(&new_rowsb[ni], 0, sizeof(MYX_RS_ROW));
        new_rowsa[ni].diff= MYX_RD_THIS_ONLY;
        new_rowsb[ni].diff= MYX_RD_OTHER_ONLY;
        ni++;
        ia++;
      }
    }
    else if (d > 0)
    {
      if (ib < last_indexb)
      {
        // this row only exists in B
        new_rowsb[ni]= rb->rows[indexb[ib]];
        memset(&new_rowsa[ni], 0, sizeof(MYX_RS_ROW));
        new_rowsa[ni].diff= MYX_RD_OTHER_ONLY;
        new_rowsb[ni].diff= MYX_RD_THIS_ONLY;
        ni++;
        ib++;
      }
    }
    else
    {
      bigint diff_mask= 0;
      int eq= 1;
      // both rows exist, compare the contents
      for (i= 0; i < ra->columns_num; i++)
      {
        if (ra->rows[indexa[ia]].fields[i].value_length != 
            rb->rows[indexb[ib]].fields[i].value_length)
        {
          eq= 0;
          diff_mask|= (1<<i);
        }
        else if (memcmp(ra->rows[indexa[ia]].fields[i].value,
                        rb->rows[indexb[ib]].fields[i].value,
                        ra->rows[indexa[ia]].fields[i].value_length)!=0)
        {
          eq= 0;
          diff_mask|= (1<<i);
        }
      }
      
      new_rowsa[ni]= ra->rows[indexa[ia]];
      new_rowsb[ni]= rb->rows[indexb[ib]];

      if (!eq)
      {
        new_rowsa[ni].diff= (diff_mask<<4) | MYX_RD_DIFFERS;
        new_rowsb[ni].diff= (diff_mask<<4) | MYX_RD_DIFFERS;
      }
      else
      {        
        new_rowsa[ni].diff= (diff_mask<<4) | MYX_RD_MATCHES;
        new_rowsb[ni].diff= (diff_mask<<4) | MYX_RD_MATCHES;
      }
      ni++;
      ia++;
      ib++;
    }
  }

  // free tmp data
  g_free(indexa);
  g_free(indexb);

  // replace rows structs from resultsets
  g_free(ra->rows);
  ra->rows= new_rowsa;
  ra->rows_num= ni;

  g_free(rb->rows);
  rb->rows= new_rowsb;
  rb->rows_num= ni;

  return 0;
}


/*
 *----------------------------------------------------------------------
 * int_compare- compare 2 strings numerically
 *
 * DESCRIPTION
 * 
 *  We compare 2 strings numerically by checking the string lengths
 * The shortest one is the smallest. If both have the same length,
 * then compare with plain strcmp. We skip all zeros at left, after
 * checking sign (and comparing the signs, if necessary).
 * 
 * RETURN VALUE
 * 
 * -1 if a < b
 * +1 if a > b
 * 0 if a == b
 *
 * NOTES
 * 
 * Only integers allowed!
 *----------------------------------------------------------------------
 */
static int int_compare(const char *a, const char *b)
{
  int asign= 1, bsign= 1;
  int alen, blen;

  // check signs
  if (*a == '-')
  {
    asign=-1;
    a++;
  }
  if (*b == '-')
  {
    bsign=-1;
    b++;
  }
  if (asign < bsign) // -x < +x
    return -1;
  if (asign > bsign) // +x > -x
    return 1;
  
  while (*a=='0') a++;
  while (*b=='0') b++;

  // check case where either is 0
  if (!*a && !*b) // both 0
    return 0;
  if (!*a) // a is 0
    return -bsign; // if b < 0, then a > b so return 1, which is -bsign
  if (!*b) // b is 0
    return asign; // if a < 0, then a < b  so return -1, which is asign
  
  alen= (int)strlen(a);
  blen= (int)strlen(b);
  
  if (alen < blen) // a < b
    return -1;
  else if (alen > blen) // a > b
    return 1;
  else
    return strcmp(a, b);
}


static int dummy_compare(const char*a, const char*b)
{
  return 0;
}




static char *replace_variables(char *str, MYX_RESULTSET *resultset)
{
  time_t t;
  char *tmp;

  tmp= str_g_replace(g_strdup(resultset->query->sql), "\r\n", " ");
  tmp= str_g_replace(tmp, "\n", " ");

  time(&t);
  str= str_g_replace(str, "$DATE$", ctime(&t));
  str= str_g_replace(str, "$QUERY$", tmp);
  g_free(tmp);

  return str;
}


int myx_export_resultset(MYX_TABLE_EXPORTER_INFO *info,
                         const char *filename,
                         const char *header_format,
                         MYX_RESULTSET *resultset)
{
  unsigned int r, c;
  unsigned int i;
  char *header;
  const char **fields;
  MYX_TE_COLUMN *columns;

  for (i= 0; i < info->option_values_num; i++)
    info->option_values[i]= replace_variables(info->option_values[i],
                                              resultset);

  columns= g_new0(MYX_TE_COLUMN, resultset->columns_num);
  for (c= 0; c < resultset->columns_num; c++)
  {
    columns[c].name= resultset->columns[c].name;
    switch (resultset->columns[c].column_type)
    {
    case MYX_RSCT_INTEGER:
      columns[c].ctype= MYX_TE_COL_INTEGER;
      break;
    case MYX_RSCT_FLOAT:
      columns[c].ctype= MYX_TE_COL_FLOAT;
      break;
    case MYX_RSCT_STRING:
    case MYX_RSCT_DATE:
    case MYX_RSCT_TIME:
    case MYX_RSCT_DATETIME:
    case MYX_RSCT_BLOB:
    case MYX_RSCT_TEXT:
    case MYX_RSCT_ENUM:
    case MYX_RSCT_SET:
      columns[c].ctype= MYX_TE_COL_STRING;
      break;
    }
  }

  if ((*info->te->setup)(info, filename, resultset->columns_num, columns) < 0)
  {
    //g_free(fields); not initialized yet
    g_free(columns);
    return -1;
  }

  (*info->te->begin)(info);
  
  header= replace_variables(g_strdup(header_format), resultset);
  (*info->te->header)(info, header);
  g_free(header);
  
  (*info->te->table_header)(info);

  fields= g_new(const char *, resultset->columns_num);
  for (r= 0; r < resultset->rows_num; r++)
  {
    unsigned int j;
    MYX_RS_ACTION *action= NULL;
    int skip= 0;

    // prepare columns
    for (c= 0; c < resultset->columns_num; c++)
      fields[c]= resultset->rows[r].fields[c].value;

    if (resultset->actions)
    {
      for (j= 0; j < resultset->actions->actions_num; j++)
      {
        if (resultset->actions->actions[j].row == r)
        {
          action= resultset->actions->actions+j;
          if (action->action == MYX_RSA_DELETE)
          {
            skip= 1;
            break;
          }
          // apply changes to this row
          if (action->action == MYX_RSA_UPDATE)
          {
            for (c= 0; c < resultset->columns_num; c++)
            {
              if (resultset->columns+c == action->column)
              {
                fields[c]= action->new_value;
                break;
              }
            }
          }
        }
      }
    }

    // check if this row was deleted
    if (!skip)
    {
      // output columns
      (*info->te->columns)(info, fields);
    }
  }

  if (resultset->actions)
  {
    int has_more_rows= 0;
    r= resultset->rows_num;
    // output remaining rows that were newly added
    do
    {
      int nothing_in_this_row= 1;
      unsigned int j;

      for (j= 0; j < resultset->columns_num; j++)
        fields[j]= NULL;

      for (j= 0; j < resultset->actions->actions_num; j++)
      {
        MYX_RS_ACTION *action= resultset->actions->actions+j;
        
        if (action->action == MYX_RSA_ADD && action->row == r)
        {
          for (c= 0; c < resultset->columns_num; c++)
          {
            if (resultset->columns+c == action->column)
            {
              fields[c]= action->new_value;
              nothing_in_this_row= 0;
              break;
            }
          }
        }
        else if (action->action == MYX_RSA_ADD)
        {
          if (action->row > r)
            has_more_rows= 1;
        }
      }
      if (!nothing_in_this_row)
      {
        // output columns
        (*info->te->columns)(info, fields);
      }
      r++;
    } while (has_more_rows);
  }

  g_free((char **)fields);
  g_free(columns);

  (*info->te->end)(info);

  return 0;
}
