/* Copyright (C) 2004 MySQL AB

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#define BUFFER_RESIZE_INCREMENT 1024*256 /* increase in 256k increments */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "myx_library.h"
#include <glib.h>
#include "myx_simple_sql_parsing.h"

/* 
 * Forward declarations
 */

static int make_state_transition(int state, bigint c);

/*
   Begin
     |
     V
 .................
 :               :<-(N)---+
 :               :        |
 :   +--(-)->[4]----(-)->[5]
 :   |        |  :
 :   |        |  :<----------------------------------------+
 :   | <------+  :                                         |
 :->[0]          :<-(!)---+      +----------+ +-(*)-+      |
 :   | <------+  :        |      |          | |     |      |
 :   |        |  :        |      V          | |     |      |
 :   |        |  :        |  +->[9]-(*)-+   | |     V      |
 :   +--(/)->[7]----(*)->[8]-+          +->[==12=====]-(/)-+
 :               :           +------(*)-+
 :               :
 :               :<-(N)---+
 :               :        |
 :               :--(#)->[6]
 :               :
 :               :        +---------+
 :               :        |         |
 :               :        V         |
 :               :<-(")->[2]-(\)->[11]
 :               :
 :               :        +---------+
 :               :        |         |
 :               :        V         |
 :               :<-(')->[3]-(\)->[10]
 :               :
 :               :<-(`)---+
 :               :        |
 :               :--(`)->[14]
 :               :
 :               :
 :               :--(;)->[1]-->[13]
 .................               |
                                 |
                                 V
                                End
*/


//N = newline, E= any other character
static int state_machine[15][12] =
{
  //     ;   '   "   #   -   /   *   N   \   E   !   `
/*0*/  { 1,  3,  2,  6,  4,  7,  0,  0,  0,  0,  0,  14 },
/*1*/  { 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13 },
/*2*/  { 2,  2,  0,  2,  2,  2,  2,  2,  11, 2,  2,  2  },
/*3*/  { 3,  0,  3,  3,  3,  3,  3,  3,  10, 3,  3,  3  },
/*4*/  { 1,  3,  2,  6,  5,  7,  0,  0,  0,  0,  0,  14 },
/*5*/  { 5,  5,  5,  5,  5,  5,  5,  0,  5,  5,  5,  5  },
/*6*/  { 6,  6,  6,  6,  6,  6,  6,  0,  6,  6,  6,  6  },
/*7*/  { 1,  3,  2,  6,  4,  7,  8,  0,  0,  0,  0,  14 },
/*8*/  { 9,  9,  9,  9,  9,  9,  12, 9,  9,  9,  0,  9  },
/*9*/  { 9,  9,  9,  9,  9,  9,  12, 9,  9,  9,  9,  9  },
/*10*/ { 3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3  },
/*11*/ { 2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2  },
/*12*/ { 9,  9,  9,  9,  9,  0,  12, 9,  9,  9,  9,  9  },
/*13*/ { 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13 },
/*14*/ { 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 0  }
};



/*
 * Public functions
 */

/**
 * This function reads the next sql-statement into
 * buffer.
 * If resize_buffer==1 then buffer is automatically
 * resized if the sql-statement is bigger than the buffer
 *
 * Buffer has to be a memory area returned by g_malloc if
 * it shall be resized!
 *
 * Return Value: the length of the next sql-statement or 0
 *
 * read_char is a function that returns the next character
 *   or -1 in case there is no more character.
 *   The first argument to read_char is arg
 *   The second argument will be set to the number of bytes read
 *   The third argument to read_char is a pointer to error-var
 *   If there is an error in read_char error should be set and 
 *   -1 be returned.
 **/
int myx_get_next_sql_statement(MYX_SQL_PARSE_ENVIRONMENT *we,
                               char **buffer, int *buffer_len,
                               int resize_buffer, int *error,
                               bigint (*read_char)(void *user_data,
                                                   int *bytes_read,
                                                   int *error),
                               void *arg)
{
  int h;
  int bytes_read= 0;
  bigint c;
  int state= we->state;

  //fast-forward
  while ( state!=1 && (c= read_char(arg, &bytes_read, error)) != -1 )
  {
    we->bytes_read+= bytes_read;
    state= make_state_transition(state, c);
  }

  // readchar returned an error
  if (*error) return 0;

  /* read in (the beginning of) the next sql-statement */
  if (state == 1)
    state= 0;

  h= 0;
  while ( state!=1 && (resize_buffer || h< *buffer_len-6) &&
         (c= read_char(arg, &bytes_read, error)) != -1 )
  {
    we->bytes_read+= bytes_read;

    if (resize_buffer && h == *buffer_len - 7)
    {
      gchar *tmp;

      /* increase buffer size */
      tmp= g_try_realloc(*buffer, *buffer_len + BUFFER_RESIZE_INCREMENT);
      if (!tmp)
      {
        *error= -1;
        return 0;
      }
      *buffer= tmp;
      *buffer_len+= BUFFER_RESIZE_INCREMENT;
    }

    if (c == '\n')
      we->line_num++;

    /* Record the character if we are in a "no comment"-section */
    if (state != 6 && state != 5 && state != 9 && state != 12)
    {
      //write
      if ( !(h==0 && g_unichar_isspace((gunichar)c)) && !(h==0 && c == 0xFEFF))
      {  /* ignore beginning white-space or a beginning BOF 
          * Windows uses a byte order mark (BOM) for example to mark a 
          *  file as UTF8
          */
        if (h == 0)
        {
          we->stmt_begin_char= we->bytes_read - 1;
          we->stmt_begin_line= we->line_num;
        }
        h+= g_unichar_to_utf8( (gunichar) c, (*buffer)+h );
      }
    }

    /* If this character triggers a comment, delete the comment-beginning */
    if (state == 4 && c == '-')
    {
      h-= 2; 
      if (h == 0)
      {
        we->stmt_begin_char= we->bytes_read - 1;
        we->stmt_begin_line= we->line_num;
      }   
    }
    else if (state == 0 && c == '#')
    {
      h-= 1;
      if (h == 0)
      {
        we->stmt_begin_char= we->bytes_read - 1;
        we->stmt_begin_line= we->line_num;
      }
    }
    else if (state == 8 && c != '!')
    {
      h-= 3;
      if (h == 0)
      {
        we->stmt_begin_char= we->bytes_read - 1;
        we->stmt_begin_line= we->line_num;
      }
    }
    state= make_state_transition(state, c);
  }

  // readchar returned an error
  if (*error) return 0;

  we->stmt_end_char= we->bytes_read - 1;
  we->stmt_end_line= we->line_num;

  (*buffer)[h]= '\0';

  we->state= state;

  return h;
}


/* Reads in the next sql-statement from file */
int myx_get_next_sql_statement_file(MYX_SQL_PARSE_ENVIRONMENT *we,
                                    MYX_INTL_FILE *stat,
                                    char **buffer, int *buffer_len,
                                    int resize_buffer, int *error)
{
  bigint (* read_char)(void *, int *,int *)= 
        (bigint (*)(void *, int *,int *)) myx_read_char_from_intl_file;
  return myx_get_next_sql_statement(we, buffer, buffer_len,
                                    resize_buffer, error,
                                    read_char,
                                    (void *)stat);
}


MYX_STRING_WRAPPER * myx_new_string_wrapper(const char *string)
{
  MYX_STRING_WRAPPER *w;

  g_return_val_if_fail( g_utf8_validate(string, -1 , NULL), NULL);

  w= g_malloc0(sizeof(MYX_STRING_WRAPPER));
  w->string= (char *)string;
  w->pos= 0;

  return w;
}

void myx_free_string_wrapper(MYX_STRING_WRAPPER *w)
{
  if (w)
  {
    g_free(w->string);
    g_free(w);
  }
}

bigint myx_read_char_from_string(void *w, int *bytes_read, int *error)
{
  MYX_STRING_WRAPPER *wrapper= (MYX_STRING_WRAPPER *) w;
  bigint c;
  bigint old_pos;

  c= g_utf8_get_char(wrapper->string + wrapper->pos);

  old_pos= wrapper->pos;

  wrapper->pos=
            g_utf8_next_char(wrapper->string + wrapper->pos) - wrapper->string;
  //the cast is ok because bytes_read can never be greater than 6
  *bytes_read= (int) (wrapper->pos - old_pos);

  return c == 0 ? EOF : c;
} 

MYX_SQL_PARSE_ENVIRONMENT* myx_new_sql_parse_environment(void)
{
  MYX_SQL_PARSE_ENVIRONMENT *we;

  we= g_malloc0(sizeof(MYX_SQL_PARSE_ENVIRONMENT));
  we->state= 1;
  we->bytes_read= 0;

  return we;
}

void myx_free_sql_parse_environment(MYX_SQL_PARSE_ENVIRONMENT *we)
{
  g_free(we);
}

/* 
 * Private functions
 */

/*
 * This function works with Unicode characters too
 * because all ASCII characters have the same code-position
 * in Unicode as in ASCII
 * It won't work on non-Ascii platforms but we don't care.
 */
static int make_state_transition(int state, bigint c)
{
  switch(c)
  {
  case ';'  : return state_machine[state][0];
  case '\'' : return state_machine[state][1];
  case '"'  : return state_machine[state][2];
  case '#'  : return state_machine[state][3];
  case '-'  : return state_machine[state][4];
  case '/'  : return state_machine[state][5];
  case '*'  : return state_machine[state][6];
  case '\n' : return state_machine[state][7];
  case '\\' : return state_machine[state][8];
  case '!'  : return state_machine[state][10];
  case '`'  : return state_machine[state][11];
  default   : return state_machine[state][9];
  }
  return state;
}

