unit MySQLConnection;

interface

uses
  gnugettext, Windows, Messages, SysUtils, Classes, ComCtrls, ImgList,
  Controls, TntComCtrls, TntClasses, Forms, Contnrs,
  ExtCtrls, AuxFuncs, PNGImage, StdCtrls,
  SyncObjs, myx_public_interface, MyxError, Options;

const
  WM_Disconnected = WM_USER+101;
  WM_Reconnected = WM_USER+102;
  WM_DefaultSchemaChanged = WM_USER+103;
  WM_SchemaListChanged = WM_USER+104;

  CR_SERVER_GONE_ERROR = 2006;
  CR_SERVER_LOST = 2013;
  CR_CONN_HOST_ERROR = 2003;

type
  TFetchDataThread = Class;

  DataBeingFetchedSet = set of 1..100;

  TThreadExecMethod = procedure(Sender: TObject) of object;

  TMySQLConn = class(TObject)
    constructor Create(StatusBar: TTntStatusBar);
    destructor Destroy; override;

    function ConnectToServer(AutoConnect: Boolean;
      MySQLConnectionsOnly: Boolean = True;
      SelectSchemata: Boolean = False): integer;

    function Reconnect: Boolean;
    procedure Disconnect;

    procedure FetchData(KindOfData: integer;
      FetchMethod: TThreadExecMethod; RefreshMethod: TThreadExecMethod;
      Target: TObject; StatusBarText: WideString;
      AllowParallelExecution: Boolean = False;
      AllowDuplicatedKindOfData: Boolean = False);
    procedure TerminateFetchingData;

    function ExecuteDirect(SQLCmd: WideString;
      DisplayErrorDialog: Boolean = True): Boolean;

    procedure SchemaListChanged;
  private
    { Private declarations }
    FDefaultSchema: WideString;

    function ShowConnectToInstanceForm(
      var User_connection: TMYX_USER_CONNECTION;
      var PMySQL: Pointer;
      MySQLConnectionsOnly: Boolean; SelectSchemata: Boolean): integer;
    function ConnectUsingUserConnection(
      Connection: WideString;
      var User_connection: TMYX_USER_CONNECTION;
      var PMySQL: Pointer; SelectSchemata: Boolean): integer;
    function ConnectUsingCommandlineConnectionParams(
      var User_connection: TMYX_USER_CONNECTION;
      var PMySQL: Pointer; SelectSchemata: Boolean): integer;
    function Connect(var PMySQL: Pointer;
      user_conn: TMYX_USER_CONNECTION;
      SelectSchemata: Boolean): integer;
  protected
    procedure SetDefaultSchema(DefaultSchema: WideString);
  public
    { Public declarations }
    //Pointer to MySQL Struct
    PMySQL: Pointer;
    MySQLMajorVersion,
    MySQLMinorVersion: integer;

    //Pointer to MYX_USER_CONNECTION
    user_connection: TMYX_USER_CONNECTION;

    ConnectionCaption: WideString;

    Params: TTntStringList;
    DoParameterSubstitution: Boolean;

    //Fetch Data Thread
    FetchDataThreadHandle: THandle;
    FetchDataThread: TFetchDataThread;

    FetchDataThreadCriticalSection: TCriticalSection;
    FetchDataMultiReadSync: TMultiReadExclusiveWriteSynchronizer;

    DataBeingFetched: DataBeingFetchedSet;
    FetchingData: Boolean;

    StatusBar: TTntStatusBar;

    Connected: Boolean;
    ConnectedToLocalhost: Boolean;

    InTransaction: Boolean;

    procedure RefreshConnectionCaption;
  published
    property DefaultSchema: WideString read FDefaultSchema write SetDefaultSchema;
  end;

  TFetchDataThread = class(TThread)
    constructor Create(MySQLConn: TMySQLConn; KindOfData: integer;
      FetchMethod: TThreadExecMethod; RefreshMethod: TThreadExecMethod;
      Target: TObject; StatusBarText: WideString; AllowParallelExecution: Boolean);
    destructor Destroy; override;
    procedure ReleaseCriticalSection;
    procedure UpdateStatusBar;
    procedure ShowError;
    procedure ExecuteSynchronized(SynchronizedMethod: TThreadExecMethod);
    procedure ExecuteSynchronizedMethod;
  protected
    procedure Execute; override;
  private
    ErrorMsg: WideString;
    FetchMethod: TThreadExecMethod;
    RefreshMethod: TThreadExecMethod;
    SynchronizedMethod: TThreadExecMethod;
  public
    MySQLConn: TMySQLConn;
    KindOfData: integer;
    Target: TObject;
    StatusBarText: WideString;
    AllowParallelExecution: Boolean;
  end;

const
  //KindOfData
  KindOfData_CatalogSchema = 1;
  KindOfData_UserNames = 2;
  KindOfData_ProcessList = 3;
  KindOfData_SchemaTables = 4;
  KindOfData_StatusVars = 5;
  KindOfData_ServerVars = 6;
  KindOfData_UserData = 7;
  KindOfData_SchemaTableStatus = 8;
  KindOfData_TableAction = 9;
  KindOfData_SchemaIndices = 10;

implementation

uses ConnectToInstance;

constructor TMySQLConn.Create(StatusBar: TTntStatusBar);
begin
  inherited Create;

  self.StatusBar:=StatusBar;

  //DLL Version check still missing

  PMySQL:=nil;
  user_connection:=nil;

  FDefaultSchema:='';

  Connected:=False;
  ConnectedToLocalhost:=False;

  //Initialize components for thread handling
  FetchDataThreadCriticalSection:=TCriticalSection.Create;
  FetchDataMultiReadSync:=TMultiReadExclusiveWriteSynchronizer.Create;
  DataBeingFetched:=[];

  FetchDataThread:=nil;
  FetchingData:=False;

  Params:=TTntStringList.Create;
  DoParameterSubstitution:=False;
end;

destructor TMySQLConn.Destroy;
begin
  if(FetchingData)then
    TerminateThread(FetchDataThreadHandle, 0);

  if(FetchDataThread<>nil)then
    FetchDataThread.Free;

  FetchDataMultiReadSync.Free;
  FetchDataThreadCriticalSection.Free;

  if(user_connection<>nil)then
    user_connection.Free;

  //Close connection and free memory for MySQL struct
  myx_mysql_close(PMySQL);

  Params.Free;

  inherited;
end;

function TMySQLConn.ConnectToServer(AutoConnect: Boolean;
  MySQLConnectionsOnly: Boolean = True;
  SelectSchemata: Boolean = False): integer;
begin
  //If AutoConnect and a connection was specified
  if(AutoConnect)and(MYXCommonOptions.ConnectionToUse<>'')then
    Result:=ConnectUsingUserConnection(MYXCommonOptions.ConnectionToUse,
      User_connection,
      PMySQL, SelectSchemata)
  //If AutoConnect and username was specified
  else if(AutoConnect)and(MYXCommonOptions.ConnectionUsername<>'')then
    Result:=ConnectUsingCommandlineConnectionParams(User_connection,
      PMySQL, SelectSchemata)
  else
    Result:=ShowConnectToInstanceForm(User_connection, PMySQL,
      MySQLConnectionsOnly, SelectSchemata);

  if(Result=1)then
  begin
    MySQLMajorVersion:=myx_get_mysql_major_version(PMySQL);
    MySQLMinorVersion:=myx_get_mysql_minor_version(PMySQL);

    RefreshConnectionCaption;

    Result:=1;

    Connected:=True;
    ConnectedToLocalhost:= (myx_is_localhost(user_connection.hostname)=1);
  end;
end;

function TMySQLConn.ShowConnectToInstanceForm(
  var User_connection: TMYX_USER_CONNECTION;
  var PMySQL: Pointer;
  MySQLConnectionsOnly: Boolean; SelectSchemata: Boolean): integer;
var ConnectToInstanceForm: TConnectToInstanceForm;
  ModalRes: integer;
begin
  ConnectToInstanceForm:=TConnectToInstanceForm.Create(nil,
    MySQLConnectionsOnly, SelectSchemata);
  try
    ModalRes:=ConnectToInstanceForm.ShowModal;
    if(ModalRes=mrOK)then
    begin
      //Initialize Variables
      User_connection:=TMYX_USER_CONNECTION.create(
        ConnectToInstanceForm.User_Connection.get_record_pointer);
      PMySQL:=ConnectToInstanceForm.PMySQL;
      DefaultSchema:=myx_get_default_schema(PMySQL);

      Result:=1;
    end
    else if(ModalRes=mrAbort)then
      Result:=-1
    else
      Result:=0;

  finally
    ConnectToInstanceForm.Free;
  end;
end;

function TMySQLConn.ConnectUsingUserConnection(
  Connection: WideString;
  var User_connection: TMYX_USER_CONNECTION;
  var PMySQL: Pointer; SelectSchemata: Boolean): integer;
var user_conns: PMYX_USER_CONNECTIONS;
  stored_conns: TMYX_USER_CONNECTIONS;
  user_conn: TMYX_USER_CONNECTION;
  error: MYX_LIB_ERROR;
  i: integer;
begin
  Result:=-1;

  //Get connection
  //Fetch connections from library
  user_conns:=myx_load_user_connections(
    MYXCommonOptions.UserDataDir+'mysqlx_user_connections.xml', @error);
  if(error<>MYX_NO_ERROR)then
    raise EMyxCommonLibraryError.Create(_('Error while loading stored connections.'),
      Ord(error), MYXCommonOptions.UserDataDir+'mysqlx_user_connections.xml');

  try
    stored_conns:=TMYX_USER_CONNECTIONS.create(user_conns);
    try
      user_conn:=nil;
      for i:=0 to stored_conns.user_connections.Count-1 do
      begin
        if(CompareText(stored_conns.user_connections[i].connection_name,
          MYXCommonOptions.ConnectionToUse)=0)then
        begin
          user_conn:=stored_conns.user_connections[i];
          break;
        end;
      end;

      //If the connection is found, connect
      if(user_conn<>nil)then
      begin
        //Copy connection
        User_connection:=TMYX_USER_CONNECTION.Create(user_conn.get_record_pointer);

        Result:=Connect(PMySQL, User_connection, SelectSchemata);
      end
      else
        ShowModalDialog(Application.Title+' - '+_('Connection not found'),
          Format(_('The connection %s cannot be found in the list of stored connections.'),
            [MYXCommonOptions.ConnectionToUse]), myx_mtError, _('OK'));
    finally
      stored_conns.Free;
    end;
  finally
    myx_free_user_connections(user_conns);
  end;
end;

function TMySQLConn.ConnectUsingCommandlineConnectionParams(
  var User_connection: TMYX_USER_CONNECTION;
  var PMySQL: Pointer; SelectSchemata: Boolean): integer;
begin
  //Create new connection form command line parameters
  User_connection:=TMYX_USER_CONNECTION.Create(
    '',
    MYXCommonOptions.ConnectionUsername,
    MYXCommonOptions.ConnectionPassword,
    MYXCommonOptions.ConnectionHost,
    StrToIntDef(MYXCommonOptions.ConnectionPort, 3306),
    MYXCommonOptions.ConnectionSchema,
    '', '', MYX_MYSQL_CONN, MYX_FAVORITE_USER_CONNECTION);

  Result:=Connect(PMySQL, User_connection, SelectSchemata);
end;

function TMySQLConn.Connect(var PMySQL: Pointer;
  user_conn: TMYX_USER_CONNECTION;
  SelectSchemata: Boolean): integer;
var ConnectionResult: integer;
begin
  //Initialize PMySQL
  PMySQL:=myx_mysql_init();
  if(PMySQL=nil)then
  begin
    ShowModalDialog(Application.Title+' '+_('Error'),
      _('Error while allocating memory for MySQL Struct.'),
      myx_mtError, 'OK');

    Result:=-1;
    Exit;
  end;

  //Connect to instance
  ConnectionResult:=myx_connect_to_instance(
    user_conn.get_record_pointer,
    PMySQL);
  if(ConnectionResult=0)then
  begin
    if(SelectSchemata)and(user_conn.schema<>'')then
    begin
      ConnectionResult:=myx_use_schema(PMySQL, user_conn.schema);
      if(ConnectionResult<>0)then
      begin
        ShowModalDialog(Application.Title+' '+_('Error'),
          Format(_('The schema %s cannot be selected.'),
            [user_conn.schema]),
          myx_mtError, 'OK');

        Result:=-1;
        Exit;
      end;

      DefaultSchema:=myx_get_default_schema(PMySQL);

    end;

    Result:=1;
  end
  else
  begin
    ShowModalDialog(Application.Title+' '+_('Error'),
      _('Could not connect to the specified host.')+' '+#13#10+#13#10+
      Format(_('MySQL Error Nr.%d'+#13#10+'%s'),
        [myx_mysql_errno(PMySQL),
        myx_mysql_error(PMySQL)]),
      myx_mtError, 'OK');

    Result:=-1;
  end;
end;

procedure TMySQLConn.RefreshConnectionCaption;
begin
  if(User_Connection<>nil)then
  begin
    ConnectionCaption:=
      User_Connection.username+'@'+
      User_Connection.hostname+':'+
      IntToStr(User_Connection.port);

    if(FDefaultSchema<>'')then
      ConnectionCaption:=ConnectionCaption+' / '+FDefaultSchema;
  end;
end;

procedure TMySQLConn.FetchData(KindOfData: integer;
  FetchMethod: TThreadExecMethod; RefreshMethod: TThreadExecMethod;
  Target: TObject; StatusBarText: WideString;
  AllowParallelExecution: Boolean;
  AllowDuplicatedKindOfData: Boolean);
begin
  if(Not(KindOfData in DataBeingFetched))or
    (AllowDuplicatedKindOfData)then
  begin
    //Add KindOfData to DataBeingFetched Set
    if(Not(KindOfData in DataBeingFetched))then
      DataBeingFetched:=DataBeingFetched + [KindOfData];
    try
      //Create a new thread
      FetchDataThread:=TFetchDataThread.Create(self,
        KindOfData, FetchMethod, RefreshMethod,
        Target, StatusBarText, AllowParallelExecution);
      try
        //Thread will be freed after execution
        FetchDataThread.FreeOnTerminate:=True;

        FetchingData:=True;

        FetchDataThread.Resume;
      except
        FetchDataThread.Free;
        FetchDataThread:=nil;

        raise;
      end;
    except
      DataBeingFetched:=DataBeingFetched - [KindOfData];

      raise;
    end;
  end;
end;

procedure TMySQLConn.TerminateFetchingData;
begin
  TerminateThread(FetchDataThreadHandle, 0);
  FetchDataThreadCriticalSection.Release;
  DataBeingFetched:=DataBeingFetched - [FetchDataThread.KindOfData];

  if(FetchDataThread<>nil)then
    FetchDataThread.Free;
  FetchDataThread:=nil;

  FetchingData:=False;

  if(StatusBar<>nil)then
  begin
    StatusBar.Panels[1].Text:='';
    StatusBar.Invalidate;
  end;
end;

function TMySQLConn.Reconnect: Boolean;
var ConnectionResult: integer;
begin
  Result:=False;

  if(User_Connection=nil)then
    Exit;

  if(PMySQL<>nil)then
    Disconnect;

  PMySQL:=myx_mysql_init();
  if(PMySQL=nil)then
    raise EMyxError.Create('Error while allocating memory for MySQL Struct.');

  ConnectionResult:=myx_connect_to_instance(
    User_Connection.get_record_pointer,
    PMySQL);
  if(ConnectionResult<>0)then
    Exit;

  Connected:=True;
  MessageToAllForms(WM_Reconnected, 0, 0);

  Result:=True;
end;

procedure TMySQLConn.Disconnect;
begin
  //Close connection and free memory for MySQL struct
  myx_mysql_close(PMySQL);
  PMySQL:=nil;

  Connected:=False;
  MessageToAllForms(WM_Disconnected, 0, 0);
end;

function TMySQLConn.ExecuteDirect(SQLCmd: WideString;
  DisplayErrorDialog: Boolean): Boolean;
var ErrorCode: MYX_LIB_ERROR;
  MySQLErrorNr: integer;
begin
  FetchDataThreadCriticalSection.Acquire;
  try
    ExecuteDirect:=False;
    try
      myx_query_execute_direct(PMySQL, SQLCmd,
        @ErrorCode);

      if(ErrorCode=MYX_NO_ERROR)then
        ExecuteDirect:=True
      else
      begin
        MySQLErrorNr:=myx_mysql_errno(PMySQL);

        if(MySQLErrorNr=CR_SERVER_GONE_ERROR)or
          (MySQLErrorNr=CR_SERVER_LOST)or
          (MySQLErrorNr=CR_CONN_HOST_ERROR)then
        begin
          Disconnect;
        end;

        if(DisplayErrorDialog)then
          ShowModalDialog('Execution Error',
            'Error while executing query.'+#13#10#13#10+
            SQLCmd+#13#10#13#10+
            'MySQL Error Nr.'+IntToStr(MySQLErrorNr)+#13#10+
            myx_mysql_error(PMySQL),
            myx_mtError, 'OK');
      end;
    except
      on x: Exception do
        if(DisplayErrorDialog)then
          ShowModalDialog('Execution Error',
            'The following error occured while executing the query.'+#13#10#13#10+
            SQLCmd+#13#10#13#10+
            x.Message, myx_mtError, 'OK');
    end;
  finally
    FetchDataThreadCriticalSection.Release;
  end;
end;

procedure TMySQLConn.SetDefaultSchema(DefaultSchema: WideString);
var error: integer;
begin
  if(PMySQL<>nil)and(DefaultSchema<>'')then
  begin
    error:=myx_use_schema(PMySQL, DefaultSchema);
    if(error=0)then
    begin
      FDefaultSchema:=myx_get_default_schema(PMySQL);

      RefreshConnectionCaption;

      MessageToAllForms(WM_DefaultSchemaChanged, 0, 0);
    end
    else
      raise EMyxSQLError.Create(Format(
        _('The default schema cannot be changed to %s'), [DefaultSchema]),
        myx_mysql_errno(PMySQL), myx_mysql_error(PMySQL));
  end;
end;

procedure TMySQLConn.SchemaListChanged;
begin
  MessageToAllForms(WM_SchemaListChanged, 0, 0);
end;

//------------------------------------------------------------------------------

constructor TFetchDataThread.Create(
  MySQLConn: TMySQLConn; KindOfData: integer;
  FetchMethod: TThreadExecMethod; RefreshMethod: TThreadExecMethod;
  Target: TObject; StatusBarText: WideString; AllowParallelExecution: Boolean);
begin
  inherited Create(True);

  self.MySQLConn:=MySQLConn;
  self.KindOfData:=KindOfData;
  self.FetchMethod:=FetchMethod;
  self.RefreshMethod:=RefreshMethod;
  self.Target:=Target;
  self.StatusBarText:=StatusBarText;
  self.SynchronizedMethod:=nil;
  self.AllowParallelExecution:=AllowParallelExecution;
end;

destructor TFetchDataThread.Destroy;
begin
  MySQLConn.FetchingData:=False;
  StatusBarText:='';
  Synchronize(UpdateStatusBar);

  MySQLConn.FetchDataThread:=nil;

  MySQLConn.DataBeingFetched:=MySQLConn.DataBeingFetched - [KindOfData];

  inherited Destroy;
end;

procedure TFetchDataThread.Execute;
begin
  //Enter CriticalSection
  if(Not(AllowParallelExecution))then
    MySQLConn.FetchDataThreadCriticalSection.Acquire;
  try
    try
      MySQLConn.FetchDataThreadHandle:=Handle;

      if(StatusBarText<>'')then
        Synchronize(UpdateStatusBar);

      //Test long response time
      //sleep(3000);

      if(Assigned(FetchMethod))then
        FetchMethod(self);

      {if(StatusBarText<>'')then
      begin
        StatusBarText:='';
        InstanceDM.FetchingData:=False;
        Synchronize(UpdateStatusBar);
      end;}

      //The Refresh Method is called Synchronized
      if(Assigned(RefreshMethod))then
        ExecuteSynchronized(RefreshMethod);
    except
      on x: EMyxSQLError do
      begin
        if(StrToIntDef(x.ErrorNr, 0)=CR_SERVER_GONE_ERROR)or
          (StrToIntDef(x.ErrorNr, 0)=CR_SERVER_LOST)or
          (StrToIntDef(x.ErrorNr, 0)=CR_CONN_HOST_ERROR)then
        begin
          MySQLConn.Disconnect;
        end;

        ErrorMsg:=x.Message+#13#10#13#10+
          'MySQL Error Nr.'+EMyxSQLError(x).ErrorNr+#13#10+
          EMyxSQLError(x).ErrorTxt;
        Synchronize(ShowError);
      end;
      on x: Exception do
      begin
        ErrorMsg:=x.Message;
        Synchronize(ShowError);
      end;
    end;
  finally
    if(Not(AllowParallelExecution))then
      MySQLConn.FetchDataThreadCriticalSection.Release;
  end;
end;

procedure TFetchDataThread.ReleaseCriticalSection;
begin
  MySQLConn.FetchDataThreadCriticalSection.Release;
end;

procedure TFetchDataThread.UpdateStatusBar;
begin
  if(MySQLConn.StatusBar<>nil)then
  begin
    if(MySQLConn.StatusBar.Panels.Count>1)then
    begin
      MySQLConn.StatusBar.Panels[1].Text:=StatusBarText;
      MySQLConn.StatusBar.Repaint;
    end;
  end;
end;

procedure TFetchDataThread.ShowError;
begin
  ShowModalDialog(Application.Title+' '+_('Error'), ErrorMsg,
    myx_mtError, _('OK'));
end;

procedure TFetchDataThread.ExecuteSynchronized(SynchronizedMethod: TThreadExecMethod);
begin
  self.SynchronizedMethod:=SynchronizedMethod;

  Synchronize(ExecuteSynchronizedMethod);
end;

procedure TFetchDataThread.ExecuteSynchronizedMethod;
begin
  if(@SynchronizedMethod<>nil)then
    SynchronizedMethod(self);
end;


end.
